from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib

from operator import itemgetter
from typing import Literal,List,Annotated
from langchain_core.messages import SystemMessage,HumanMessage,AIMessage,BaseMessage,ToolMessage
from langchain_core.pydantic_v1 import BaseModel
from typing_extensions import TypedDict
from langchain_core.tools import tool,Tool
from langchain_core.runnables import RunnablePassthrough,RunnableLambda,RunnableParallel
from langgraph.graph import START,END,StateGraph,MessagesState
from langgraph.checkpoint.sqlite import SqliteSaver

from langgraph.prebuilt import ToolNode, tools_condition
from langgraph.graph.state import CompiledStateGraph

from ylz_utils.file import FileLib
from ylz_utils.data import StringLib,Color
from ylz_utils.langchain.graph.stand_graph import StandGraph



class GraphLib():
    def __init__(self,langchainLib:LangchainLib,db_conn_string=":memory:"):
        self.langchainLib = langchainLib
        #self.websearch_tool = self.set_websearch_tool(websearch_key)        
        #self.ragsearch_tool = langchainLib.get_ragsearch_tool(retriever)
        self.websearch_tool = None
        self.ragsearch_tool = None
        self.python_repl_tool = langchainLib.get_python_repl_tool()
        self.memory = SqliteSaver.from_conn_string(db_conn_string)
        self.stand_graph = StandGraph(self)

    def set_dbname(self,dbname):
        # "checkpoint.sqlite"
        self.memory = SqliteSaver.from_conn_string(dbname)
    def set_ragsearch_tool(self,retriever):
        name = "rag_searcher"
        description = "一个有用的工具用来从本地知识库中获取信息。你总是利用这个工具优先从本地知识库中搜索有用的信息"
        self.ragsearch_tool = self.langchainLib.get_ragsearch_tool(retriever,name,description)
    def set_websearch_tool(self,websearch_key):
        self.websearch_tool = self.langchainLib.get_websearch_tool(websearch_key)
    
    def create_response(self, response: str, ai_message: AIMessage):
        return ToolMessage(
            content = response,
            tool_call_id = ai_message.tools_calls[0]["id"]
        )
         
    def get_graph(self,llm_key=None,llm_model=None,key:Literal['stand_graph']='stand_graph',):
        if key=='other_graph':
            pass
        else:
            return self.stand_graph.get_graph(llm_key=llm_key,llm_model=llm_model)

    def graph_stream(self,graph,message,thread_id="default-default",system_message=None):    
        messages = []
        if system_message:
            messages.append(SystemMessage(content=system_message))
        if message:
            messages.append(HumanMessage(content=message))
            values = {"messages":messages}          
        else:
            values = None
        events = graph.stream( values,
                                config = {"configurable":{"thread_id":thread_id}},
                                stream_mode = "values")
        _print_set = set()
        for event in events:
            self._print_event(event,_print_set)
    
    def _print_event(self, event: dict, _printed: set, max_length=1500):
        current_state = event.get("dialog_state")
        if current_state:
            print("Currently in: ", current_state[-1])
        message = event.get("messages")
        if message:
            if isinstance(message, list):
                message = message[-1]
            if message.id not in _printed:
                msg_repr = message.content
                if len(msg_repr) > max_length:
                    msg_repr = msg_repr[:max_length] + " ... (truncated)"
                if isinstance(message,AIMessage):
                    if message.tool_calls:
                        print(f"{Color.LBLUE}AI:{Color.RESET}",f'使用{Color.GREEN}{message.tool_calls[0]["name"]}{Color.RESET},调用参数:{Color.GREEN}{message.tool_calls[0]["args"]}{Color.RESET}')
                    else:
                        print(f"{Color.LBLUE}AI:{Color.RESET}",msg_repr,
                              f'[model:{Color.LYELLOW}{message.response_metadata.get("model_name")}{Color.RESET},token:{Color.LYELLOW}{message.usage_metadata["total_tokens"]}{Color.RESET}]')
                elif isinstance(message,ToolMessage):
                    print(f"    {Color.BLUE}Tool:{Color.RESET}",msg_repr)
                elif isinstance(message,HumanMessage):
                    print(f"{Color.BLUE}User:{Color.RESET} {msg_repr}")
                _printed.add(message.id)
                
    def graph_get_state_history(self,graph:CompiledStateGraph,thread_id="default-default"):
        state_history = graph.get_state_history(config = {"configurable":{"thread_id":thread_id}} )
        for state in state_history:
            print("Num Messages: ", len(state.values["messages"]), "Next: ", state.next)
            print("-" * 80)
        return state_history
    
    def graph_get_state(self,graph:CompiledStateGraph,thread_id="default-default"):
        state = graph.get_state(config =  {"configurable":{"thread_id":thread_id}})
        return state
    
    def graph_update_state(self,graph:CompiledStateGraph,thread_id,values,as_node = None):
        if not as_node:
            graph.update_state(config = {"configurable":{"thread_id":thread_id}},values=values)

    def export_graph(self,graph:CompiledStateGraph):
        FileLib.writeFile("graph.png",graph.get_graph().draw_mermaid_png(),mode="wb")    



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib

from langchain_community.document_loaders.image import UnstructuredImageLoader

class ImageLib():
     def __init__(self,langchainLib:LangchainLib):
          self.langchainLib = langchainLib
     def loader(self,filename):
         "./example_data/layout-parser-paper-screenshot.png"
         return UnstructuredImageLoader(filename)


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib

from langchain_community.utilities.wolfram_alpha import WolframAlphaAPIWrapper

class WolframAlphaTool():
    def __init__(self,langchainLib:LangchainLib):
        self.langchainLib = langchainLib
        self.config = langchainLib.config
    def get_tool(self):
        WOLFRAM_ALPHA_APPID = "U83YPW-GR2VEAQJV5"
        wolfram = WolframAlphaAPIWrapper(wolfram_alpha_appid=WOLFRAM_ALPHA_APPID)
        return wolfram



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib

from ylz_utils.langchain.vectorstores.elasticsearch import ESLib
from ylz_utils.langchain.vectorstores.faiss import FaissLib


class VectorstoreLib():
    def __init__(self,langchainLib:LangchainLib):
        self.faissLib = FaissLib(langchainLib)
        self.esLib = ESLib(langchainLib)


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib

from langchain_community.document_loaders import Docx2txtLoader
from langchain_text_splitters import RecursiveCharacterTextSplitter
class DocxLib():
    def __init__(self,langchainLib:LangchainLib):
        self.langchainLib = langchainLib
    def loader(self, filename: str):
        #"./example_data/fake.docx"
       return Docx2txtLoader(filename)
    def load_and_split(self, docx_file, chunk_size=1000,chunk_overlap=0):
        loader = self.loader(docx_file)
        docs = loader.load()
        spliter:RecursiveCharacterTextSplitter = self.langchainLib.get_textsplitter(chunk_size=chunk_size,chunk_overlap=chunk_overlap)
        splited_docs = spliter.split_documents(docs)
        return splited_docs
    


from ylz_utils.langchain import LangchainLib

def input_with_readline(prompt):
    try:
        return input(prompt)
    except UnicodeDecodeError:
        print("输入的内容存在编码问题，请确保使用 UTF-8 编码的字符，并不要使用回退键。")
        return input_with_readline(prompt)
    
def start_chat(langchainLib:LangchainLib,args):
    llm_key = args.llm_key
    message = args.message
    model = args.llm_model
    user_id = args.user if args.user else 'default'
    conversation_id = args.conversation if args.conversation else 'default'
    llm = langchainLib.get_llm(key=llm_key,model=model)
    dbname = args.chat_dbname or 'chat.sqlite'
    #### chat 模式
    prompt = langchainLib.get_prompt(use_chat=True)
    if dbname:
        langchainLib.llmLib.set_dbname(dbname)
    chat = langchainLib.get_chat(llm,prompt)
    chain = chat | langchainLib.get_outputParser()
    while True:
        if not message:
            message=input_with_readline("USER: ")
        else:
            print(f"USER: {message}")
        if message.lower() in ['/quit','/exit','/stop','/bye','/q']:
            print("Good bye!!")
            break
        res = chain.stream({"input":message}, {'configurable': {'user_id': user_id,'conversation_id': conversation_id}} )
        message = ""
        print("AI: ",end="")
        for chunk in res:
            print(chunk,end="")
        print("\n")
    print("\n",langchainLib.get_llm(full=True))


from ylz_utils.langchain import LangchainLib
from langchain_core.messages import HumanMessage,ToolMessage

def input_with_readline(prompt):
    try:
        return input(prompt)
    except UnicodeDecodeError:
        print("输入的内容存在编码问题，请确保使用 UTF-8 编码的字符，并不要使用回退键。")
        return input_with_readline(prompt)
    
def start_graph(langchainLib:LangchainLib,args):
    # # 设置标准输入为 UTF-8 编码
    # sys.stdin = codecs.getreader('utf-8')(sys.stdin.detach())

    llm_key = args.llm_key
    llm_model = args.llm_model
    message = args.message
    chat_dbname = args.chat_dbname
    rag_dbname = args.rag_dbname
    user = args.user or 'default'
    conversation = args.conversation or 'default'
    thread_id = f"{user}-{conversation}"
    websearch_key = args.websearch
    if chat_dbname:                                  
        langchainLib.graphLib.set_dbname(chat_dbname)
        print("!!!",f"使用对话数据库{chat_dbname}")
    if rag_dbname:
        retriever = langchainLib.vectorstoreLib.faiss.load(rag_dbname).as_retriever()
        langchainLib.graphLib.set_ragsearch_tool(retriever)
        print("!!!",f"使用知识库{rag_dbname}")
    if websearch_key:
        langchainLib.graphLib.set_websearch_tool(websearch_key)
        print("!!!",f"使用搜索工具{websearch_key}")
        
    graph = langchainLib.get_graph(llm_key=llm_key,llm_model=llm_model)
    system_message = \
"""
请始终使用中文，并确保中文正确。
要求如下：
 - 一步一步解决问题，过程中进行反思以达到最佳回复
 - 只有计算和时间问题才使用python_repl工具,使用python_repl工具时要记得用print函数返回结果。陈述事实的时候请不要使用工具
 - 不要产生幻觉，不知道的问题优先从互联网查询，关于用户自己的问题可以向用户询问。
 - 当碰到需要需要用户来确认的问题或你需要用户告诉你的问题时，请使用使用human工具向用户询问。注意，询问时请提出询问的具体问题，不要重复我提出的问题
"""
    while True:
        if not message:
            #message = input("User Input: ")
            message = input_with_readline("User Input: ")
        else:
            print(f"User:{message}")
        if message.lower() in ["/quit", "/exit", "/stop","/q","/bye"]:
            print("Goodbye!")
            break
        #system_message = """请始终用中文回答"""
        langchainLib.graphLib.graph_stream(graph,message,thread_id = thread_id,system_message=system_message)
        system_message=None
        snapshot = langchainLib.graphLib.graph_get_state(graph,thread_id)
        if snapshot.next: 
            question = snapshot.values["messages"][-1].tool_calls[0]["args"]["request"]            
            tool_call_id = snapshot.values["messages"][-1].tool_calls[0]["id"]
            message = input_with_readline("User Input: ")
            if message.strip().lower() != '':
                tool_message = ToolMessage(tool_call_id=tool_call_id, content=message)
                langchainLib.graphLib.graph_update_state(graph,thread_id=thread_id, values = {"messages":[tool_message]})
                langchainLib.graphLib.graph_stream(graph,None,thread_id = thread_id)

        message = ""

    current_state = langchainLib.graphLib.graph_get_state(graph,thread_id)
    print("\n本次对话的所有消息:\n",current_state.values["messages"])

    langchainLib.graphLib.export_graph(graph)



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib

from langchain_elasticsearch import ElasticsearchStore,DenseVectorStrategy
from langchain_elasticsearch.client import create_elasticsearch_client
from langchain_core.documents import Document
from tqdm import tqdm
from typing import List


from ylz_utils.config import Config

class ESLib():
    def __init__(self,langchainLib:LangchainLib):
        self.config = Config()
        self.langchainLib = langchainLib
        self.client = None
        self.es_host=self.config.get("ES.HOST")
        self.es_user=self.config.get("ES.USER")
        self.es_password=self.config.get("ES.PASSWORD")
    def init_client(self,host=None,es_user=None,es_password=None,connect_string:str=None):
        if connect_string:
            if connect_string.startswith("es:///"):
                es_url = self.config.get("ES.HOST")
                es_username = self.config.get("ES.USER")
                es_password = self.config.get("ES.PASSWORD")
                index_name = connect_string.split(':///')[1]
            else:
                es_url = f"https://{connect_string.split('@')[1].split('/')[0]}"
                es_username = connect_string.split('@')[0].split('://')[1].split(":")[0]
                es_password = connect_string.split('@')[0].split('://')[1].split(":")[1]
                index_name = connect_string.split('@')[1].split('/')[1]
            es_connection = create_elasticsearch_client(
                url=es_url or self.es_host,
                username=es_username or self.es_user,
                password=es_password or self.es_password,
                params = {"verify_certs":False,"ssl_show_warn":False},
            )
        else:
            es_connection = create_elasticsearch_client(
                url=host or self.es_host,
                username=es_user or self.es_user,
                password=es_password or self.es_password,
                params = {"verify_certs":False,"ssl_show_warn":False},
            )
            index_name = None
        self.client = es_connection
        return self.client,index_name
    
    def get_store(self,index_name="langchain_index",embedding=None) -> ElasticsearchStore:        
        if not embedding:
            embedding = self.langchainLib.get_embedding()
        return ElasticsearchStore(
                    es_connection=self.client,
                    index_name=index_name,
                    embedding=embedding
                )
    def delete_store(self,index_name):
        # "test-metadata, test-elser, test-basic"
        return self.client.indices.delete(
            index=index_name,
            ignore_unavailable=True,
            allow_no_indices=True
        )
            
    def create_from_docs(self,vectorstore:ElasticsearchStore,docs) -> List[str]:
        all_ids = []
        with tqdm(total= len(docs)) as pbar:
            for idx,doc in enumerate(docs):
                if doc.page_content:
                    ids = vectorstore.add_documents([doc])
                    all_ids.extend(ids)
                pbar.update(1)
        return all_ids
    def create_from_texts(self,vectorstore:ElasticsearchStore,texts) -> List[str]:
        all_ids = []
        with tqdm(total= len(texts)) as pbar:
            for idx,text in enumerate(texts):
                if text:
                    ids = vectorstore.add_texts([text])
                    all_ids.extend(ids)
                pbar.update(1)
        return all_ids
        
    def delete(self,vectorstore: ElasticsearchStore,ids: List[str] | None = None):
        return vectorstore.delete(ids) 
          
    def search(self,query,vectorstore: ElasticsearchStore,k=10,filter=[]):
        return vectorstore.similarity_search(query,k=k,filter=filter)
    
    def search_with_score(self,query,vectorstore: ElasticsearchStore,k=10,filter=[]):
        return vectorstore.similarity_search_with_score(query,k=k,filter=filter)



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib

#from langchain_community.document_loaders import UnstructuredPowerPointLoader
from langchain_core.documents import Document
from langchain_text_splitters import RecursiveCharacterTextSplitter
import pptx 
from pptx.slide import Slide,SlideLayout,SlideShapes
from pptx.presentation import Presentation
from typing import List,Dict, Literal, Tuple
from pptx.shapes.autoshape import Shape, MSO_SHAPE_TYPE
from pptx.enum.shapes import MSO_SHAPE
from pptx.chart.data import ChartData,XyChartData,BubbleChartData
from pptx.dml.color import RGBColor
from pptx.enum.chart import XL_CHART_TYPE
from pptx.enum.chart import XL_LEGEND_POSITION
from pptx.enum.text import PP_ALIGN
from pptx.util import Inches,Cm,Pt
class PptxLib():
    ppt: Presentation  = None
    slides: List[Slide] = []
    unit = None
    def __init__(self,langchainLib:LangchainLib):
        self.langchainLib = langchainLib
        self.unit = Pt
    def set_default_unit(self,unit):
        if unit.lower()=="inches":
            self.unit = Inches
        elif unit.lower()=="cm":
            self.unit = Cm
        elif unit.lower()=="pt":
            self.unit = Pt
        else:
            raise Exception("必须指定Inches,Cm,Pt中的一种单位")
    def _get_sizes(self,left:float|int,top:float|int,width:float|int,height:float|int):
        if self.unit == Inches:
            return Inches(left),Inches(top),Inches(width),Inches(height)
        elif self.unit == Cm:
            return Cm(left),Cm(top),Cm(width),Cm(height)
        else:
            return Pt(left),Pt(top),Pt(width),Pt(height)
    def _get_size(self,size:float|int):
        if self.unit == Inches:
            return Inches(size)
        elif self.unit == Cm:
            return Cm(size)
        else:
            return Pt(size)
    def newer(self,filename:str):
        self.ppt = pptx.Presentation()
        self.filename = filename
        return self
    def loader(self, filename: str):
        #"./example_data/fake-power-point.pptx"
        self.ppt = pptx.Presentation(filename)
        self.filename = filename
        return self
    def lazy_load(self):
        if not self.ppt:
            raise Exception("没有正确指定ppt loader!")
        for slide in self.ppt.slides:
            text = ""
            for shape in slide.shapes:
                if shape.has_text_frame:
                    text_frame = shape.text_frame
                    text = text + text_frame.text + "\n"
            yield Document(text)
    def load(self):
        docs = []
        for doc in self.lazy_load():
            docs.append(doc)
        return docs

    def load_and_split(self, docx_file, chunk_size=1000,chunk_overlap=0):
        loader = self.loader(docx_file)
        docs = loader.load()
        spliter:RecursiveCharacterTextSplitter = self.langchainLib.get_textsplitter(chunk_size=chunk_size,chunk_overlap=chunk_overlap)
        splited_docs = spliter.split_documents(docs)
        return splited_docs
    def add_slide(self,slide_layout=None):
        if not slide_layout:
            slide_layout = self.ppt.slide_layouts[1]
        else:
            slide_layout = self.ppt.slide_layouts[slide_layout]
        slide = self.ppt.slides.add_slide(slide_layout)
        self.slides.append(slide)
        return self
    def set_title(self,title,subtitle=None,slide_idx=None):
        if slide_idx:
            slide = self.slides[slide_idx]
        else:
            slide = self.slides[-1]
        shapes = slide.shapes
        shapes.title.text = title
        if subtitle:
            slide.placeholders[1].text = subtitle
        return self
    def add_text(self,text,left,top,width,height,
                 word_wrap=False,
                 name=None,
                 bold=None,
                 italic=None,
                 size=None,
                 href=None,
                 alignment=None,
                 color=None,
                 slide_idx=None):
        if slide_idx:
            slide = self.slides[slide_idx]
        else:
            slide = self.slides[-1]
        left,top,width,height = self._get_sizes(left,top,width,height)
        shapes = slide.shapes
        txBox = shapes.add_textbox(left,top,width,height)
        tf = txBox.text_frame
        if word_wrap:
            tf.word_wrap = True
        if isinstance(text,list):
            tf.text = text[0]
            for rest_text in text[1:]:
                self.add_text_paragraph(tf,rest_text,
                                        name=name,
                                        bold=bold,
                                        italic=italic,
                                        size=size,
                                        href=href,
                                        alignment=alignment,
                                        color=color,
                                        level=1)
        else:
            tf.text = text
        return tf
    def add_text_paragraph(self,tf,text,
                           name=None,bold=False,italic=False,size=20,
                           href=None,
                           alignment=None,
                           color:Tuple[int,int,int]=None,level=0):
        p=tf.add_paragraph()
        p.text = text
        if name:
            p.font.name = name
        p.font.blod = bold
        p.font.italic = italic
        p.font.size = Pt(size)
        if alignment:
            if alignment.lower()=="left":
                p.alignment = PP_ALIGN.LEFT
            elif alignment.lower()=="right":
                p.alignment = PP_ALIGN.RIGHT
            elif alignment.lower()=="center":
                p.alignment = PP_ALIGN.CENTER
        if color:
            p.font.color.rgb = RGBColor(*color)
        if href:
            p.hyperlink.address = href
        p.level = level
        return p
    def add_image(self,image_path,left,top,width,height,slide_idx=None):
        if slide_idx:
            slide = self.slides[slide_idx]
        else:
            slide = self.slides[-1]
        left,top,width,height = self._get_sizes(left,top,width,height)
        shapes = slide.shapes
        image = shapes.add_picture(image_path,left,top,width,height)
        return image
    def add_chart(self,chart_type:Literal['pie','column','scatter','bubble'],
                  data:List[Dict[str,Dict]] | List[Dict[str,List[Tuple[int,int]]]] | List[Dict[str,List[Tuple[int,int,int]]]],
                  x,y,cx,cy,
                  smooth:List[bool]=None,
                  label = False,
                  label_size = None,
                  label_color: Tuple[int,int,int] = None,
                  legend = True,
                  legend_position:Literal['top','left','right','bottom'] = None,
                  slide_idx=None):
        if slide_idx:
            slide = self.slides[slide_idx]
        else:
            slide = self.slides[-1]
        x,y,cx,cy = self._get_sizes(x,y,cx,cy)
        shapes = slide.shapes
        if chart_type.lower() == 'pie':
            chart_type = XL_CHART_TYPE.PIE
        elif chart_type.lower() == 'column':
            chart_type = XL_CHART_TYPE.COLUMN_CLUSTERED
        elif chart_type.lower() == 'scatter':
            chart_type = XL_CHART_TYPE.XY_SCATTER
        elif chart_type.lower() == 'bubble':
            chart_type = XL_CHART_TYPE.BUBBLE
        elif chart_type.lower() == 'line':
            chart_type = XL_CHART_TYPE.LINE
        else:
            chart_type = XL_CHART_TYPE.COLUMN_CLUSTERED

        
        if chart_type == XL_CHART_TYPE.XY_SCATTER:
            chart_data = XyChartData()
            for idx,item in enumerate(data):
                s = chart_data.add_series(list(item.keys())[0])
                for value in list(item.values())[0]:
                    s.add_data_point(*value)
        elif chart_type == XL_CHART_TYPE.BUBBLE:
            chart_data = BubbleChartData()
            for idx,item in enumerate(data):
                s = chart_data.add_series(list(item.keys())[0])
                for value in list(item.values())[0]:
                    s.add_data_point(*value)
        else:
            chart_data = ChartData()
            categories = list(list(data[0].values())[0].keys())
            chart_data.categories = categories
            for idx,item in enumerate(data):
                chart_data.add_series(list(item.keys())[0],list(list(item.values())[0].values()))
        chart_frame = shapes.add_chart(chart_type,x,y,cx,cy,chart_data)
        chart = chart_frame.chart
        if smooth and chart_type.lower == 'line' and len(smooth)==len(data):
            for idx,is_smooth in enumerate(smooth):
                chart_data.series[idx].smooth = is_smooth
        if label:
            plot = chart.plots[0]
            plot.has_data_labels = True
            data_labels = plot.data_labels
            if label_size:
                data_labels.font.size = self._get_size(label_size)
            if label_color:
                data_labels.font.color.rgb = RGBColor(*label_color)
        if legend:
            chart.has_legend = True
            if legend_position:
                if legend_position.lower() == 'left':
                    chart.legend.position = XL_LEGEND_POSITION.LEFT
                elif legend_position.lower() == 'right':
                    chart.legend.position = XL_LEGEND_POSITION.RIGHT
                elif legend_position.lower() == 'top':
                    chart.legend.position = XL_LEGEND_POSITION.TOP
                elif legend_position.lower() == 'bottom':
                    chart.legend.position = XL_LEGEND_POSITION.BOTTOM
        return chart
    
    def add_shape(self,shape_type,left,top,width,height,text=None,
                  fill_type:Literal["solid","transparent"]|None=None,
                  fill_color=None,
                  line_color=None,
                  line_brightness=None,
                  line_width=None,
                  rotation=None,
                  slide_idx=None) -> Shape:
        if slide_idx:
            slide = self.slides[slide_idx]
        else:
            slide = self.slides[-1]
        left,top,width,height = self._get_sizes(left,top,width,height)
        shapes = slide.shapes
        shape = shapes.add_shape(shape_type,left,top,width,height)
        shape.text = text
        if fill_type:
            if fill_type=="solid":
                shape.fill.solid()
            elif fill_type=="transparent":
                shape.fill.background()
        if fill_color and fill_type!="transparent":
            shape.fill.color.rgb = RGBColor(*fill_color)
        if line_color: 
            shape.line.color.rgb = RGBColor(*line_color)
        if line_brightness:
            shape.line.brightness = line_brightness
        if line_width:
            shape.line.width = self._get_size(line_width)
        if rotation:
            shape.rotation = rotation
        print(type(shape))
        return shape
    def get_shape_sizes(self,shape:Shape,unit=None):
        if not unit:
            unit = self.unit
        if unit==Inches:
            return shape.left.inches,shape.top.inches,shape.width.inches,shape.height.inches
        elif unit == Cm:
            return shape.left.cm,shape.top.cm,shape.width.cm,shape.height.cm
        else:
            return shape.left.pt,shape.top.pt,shape.width.pt,shape.height.pt 
    def add_table(self,contents:List[Dict],left,top,width,height,slide_idx=None,with_header=False):
        if slide_idx:
            slide = self.slides[slide_idx]
        else:
            slide = self.slides[-1]
        left,top,width,height = self._get_sizes(left,top,width,height)
        shapes = slide.shapes
        new_contents = contents.copy()
        if len(new_contents)==0:
            raise Exception("表格应至少包含一行一列!")
        if with_header:
            header = {}
            for key in new_contents[0]:
                header[key] = key
            new_contents.insert(0,header)
        
        rows = len(new_contents)
        cols = len(new_contents[0].keys())
        
        table = shapes.add_table(rows,cols,left,top,width,height).table
        for idx,row in enumerate(new_contents):
            for jdx,col_key in enumerate(row):
                table.cell(idx,jdx).text = str(new_contents[idx][col_key])
        return table
    def get_slide_master(self):
        slide_master = self.ppt.slide_master
        return slide_master
    def save(self):
        self.ppt.save(self.filename)



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib
    
from langchain_core.output_parsers import StrOutputParser,JsonOutputParser
from langchain_core.output_parsers import PydanticOutputParser
from langchain.output_parsers import OutputFixingParser
from langchain.output_parsers import RetryOutputParser

from langchain_core.prompts import PromptTemplate

from langchain_core.pydantic_v1 import BaseModel, Field, validator
from typing import List,Literal

class OutputParserLib():
   def get_outputParser(self,pydantic_object=None,fix=False,llm=None,retry=1):
        NAIVE_FIX = """Instructions:
            --------------
            {instructions}
            --------------
            Completion:
            --------------
            {input}
            --------------

            Above, the Completion did not satisfy the constraints given in the Instructions.
            Error:
            --------------
            {error}
            --------------

            Please try again. Please only respond with an answer that satisfies the constraints laid out in the Instructions:"""

        PROMPT = PromptTemplate.from_template(NAIVE_FIX)
        
        if pydantic_object:
            parser =  PydanticOutputParser(pydantic_object=pydantic_object)
        else:
            parser = StrOutputParser()
        if fix:
            if not llm:
                llm = self.get_llm()
            OutputFixingParser.legacy = False
            parser =  OutputFixingParser.from_llm(
                llm = llm,
                prompt = PROMPT ,
                parser = parser,
                max_retries = retry
            )
        return parser 


from fastapi import FastAPI
from langserve import add_routes
import uvicorn

from ylz_utils.langchain import LangchainLib

def serve(args):
    print("args:",args)
    langchainLib: LangchainLib = LangchainLib()
    langchainLib.add_plugins()    
    path = args.path
    host = args.host
    port = args.port
    llm_key = args.llm
    model = args.model
    app = FastAPI(title="Langserve")
    chain = langchainLib.get_prompt(human_keys={"input":"问题"}) | langchainLib.get_llm(llm_key,model) | langchainLib.get_outputParser()
    add_routes(app,runnable=chain,path=path)
    uvicorn.run(app, host = host, port = port)



from langchain_core.messages import ToolMessage
from langchain_core.runnables import RunnableLambda

from langgraph.prebuilt import ToolNode


def handle_tool_error(state) -> dict:
    error = state.get("error")
    tool_calls = state["messages"][-1].tool_calls
    return {
        "messages": [
            ToolMessage(
                content=f"Error: {repr(error)}\n please fix your mistakes.",
                tool_call_id=tc["id"],
            )
            for tc in tool_calls
        ]
    }


def create_tool_node_with_fallback(tools: list) -> dict:
    return ToolNode(tools).with_fallbacks(
        [RunnableLambda(handle_tool_error)], exception_key="error"
    )


def _print_event(event: dict, _printed: set, max_length=1500):
    current_state = event.get("dialog_state")
    if current_state:
        print("Currently in: ", current_state[-1])
    message = event.get("messages")
    if message:
        if isinstance(message, list):
            message = message[-1]
        if message.id not in _printed:
            msg_repr = message.pretty_repr(html=True)
            if len(msg_repr) > max_length:
                msg_repr = msg_repr[:max_length] + " ... (truncated)"
            print(msg_repr)
            _printed.add(message.id)


from __future__ import annotations
from typing import TYPE_CHECKING,Literal,TypedDict,Annotated
if TYPE_CHECKING:
    from ylz_utils.langchain.graph import GraphLib

from langgraph.graph.state import CompiledStateGraph
from langchain_core.pydantic_v1 import BaseModel
from langgraph.graph import START,END,StateGraph,MessagesState
from langchain_core.messages import SystemMessage,HumanMessage,AIMessage,BaseMessage,ToolMessage


from langgraph.prebuilt import ToolNode, tools_condition
from langgraph.graph.message import add_messages

from langgraph.graph.state import CompiledStateGraph

class State(TypedDict):
    messages: Annotated[list,add_messages]
    ask_human: bool

class RequestAssistance(BaseModel):
    """Escalate the conversation to an expert. Use this if you are unable to assist directly or if the user requires support beyond your permissions.

    To use this function, relay the user's 'request' so the expert can provide the right guidance.
    """
    request: str   

class StandGraph():
    def __init__(self,graphLib:GraphLib):
        self.graphLib = graphLib
        self.llm_with_tools = None

    def chatbot(self,state:State):
        response = self.llm_with_tools.invoke(state["messages"])
        ask_human = False
        if response.tool_calls and response.tool_calls[0]["name"] == RequestAssistance.__name__:
            ask_human = True
        return {"messages":[response],"ask_human":ask_human}
    
    def human_node(self, state: State):
        new_messages = []
        if not isinstance(state["messages"][-1],ToolMessage):
            new_messages.append(
                self.create_response("No response from human.",state["messages"][-1])
            )
        return {
            "messages": new_messages,
            "ask_human": False
        }
    
    def select_next_node(self, state:State) -> Literal["human","tools","__end__"]:
        if state["ask_human"]:
            return "human"
        return tools_condition(state)

    def translate_to_en(self,state:State):
        return state
        content = state["messages"][-1].content
        prompt = self.langchainLib.get_prompt(
"""
将以下句子翻译成英文,注意不要遗漏任何信息，数字不是整数的话要保留所有整数和小数，不要翻译任何公司或人名
然后一步一步获得最终答案
""",
        use_chat=False)
        #llm = self.langchainLib.get_llm(model = "llama3-groq-70b-8192-tool-use-preview")
        #llm = self.langchainLib.get_llm(model = "llama3-70b-8192")
        llm = self.langchainLib.get_llm(key = "LLM.TOGETHER")
        chain = prompt | llm
        response = chain.invoke({"input":content})
        return {"messages":[response],"ask_human":False}

    def router(self,state:MessagesState)-> Literal["tools","__end__"]:
        messages = state["messages"]
        tool_calls = messages[-1].tool_calls
        if tool_calls:
            return "tools"
        else:
            return END
    def get_graph(self,llm_key=None,llm_model=None) -> CompiledStateGraph:
        llm = self.graphLib.langchainLib.get_llm(llm_key,llm_model)
        tools = [
            self.graphLib.python_repl_tool,
        ]
        if self.graphLib.websearch_tool:
            tools.append(self.graphLib.websearch_tool)
        if self.graphLib.ragsearch_tool:
            tools.append(self.graphLib.ragsearch_tool)
        # tools = [
        #        Tool(name="websearch",
        #             description = "the tool when you can not sure,please search from the internet",
        #             func = websearch
        #        ),
        #        Tool(name="python_repl",
        #             description="when you need to calculation, use python repl tool to execute code ,then return the result to AI.",
        #             func=self.python_repl_tool)
        #      ]
        self.llm_with_tools = llm.bind_tools(tools+[ RequestAssistance ])      

        workflow = StateGraph(State)
        workflow.add_node("translate",self.translate_to_en)
        workflow.add_node("chatbot",self.chatbot)
        workflow.add_node("tools",ToolNode(tools=tools))
        workflow.add_node("human", self.human_node)
        
        workflow.add_conditional_edges("chatbot", self.select_next_node,
                                       {"human":"human","tools":"tools","__end__":"__end__"})
        
        workflow.add_edge(START, "translate")
        workflow.add_edge("translate","chatbot")
        workflow.add_edge("tools","chatbot")
        workflow.add_edge("human","chatbot")

        graph = workflow.compile(checkpointer=self.graphLib.memory,
                                 interrupt_before= ["human"])

        return graph    



import shutil
import uuid

from ylz_utils.cli.custom_service.db_init import backup_file, db
from ylz_utils.cli.custom_service.graph import part_1_graph
from ylz_utils.cli.custom_service.utils import _print_event

# Let's create an example conversation a user might have with the assistant
tutorial_questions = [
    "Hi there, what time is my flight?",
    # "Am i allowed to update my flight to something sooner? I want to leave later today.",
    # "Update my flight to sometime next week then",
    # "The next available option is great",
    # "what about lodging and transportation?",
    # "Yeah i think i'd like an affordable hotel for my week-long stay (7 days). And I'll want to rent a car.",
    # "OK could you place a reservation for your recommended hotel? It sounds nice.",
    # "yes go ahead and book anything that's moderate expense and has availability.",
    # "Now for a car, what are my options?",
    # "Awesome let's just get the cheapest option. Go ahead and book for 7 days",
    # "Cool so now what recommendations do you have on excursions?",
    # "Are they available while I'm there?",
    # "interesting - i like the museums, what options are there? ",
    # "OK great pick one and book it for my second day there.",
]

# Update with the backup file so we can restart from the original place in each section
shutil.copy(backup_file, db)
thread_id = str(uuid.uuid4())

config = {
    "configurable": {
        # The passenger_id is used in our flight tools to
        # fetch the user's flight information
        "passenger_id": "3442 587242",
        # Checkpoints are accessed by thread_id
        "thread_id": thread_id,
    }
}


_printed = set()
for question in tutorial_questions:
    events = part_1_graph.stream(
        {"messages": ("user", question)}, config, stream_mode="values"
    )
    for event in events:
        _print_event(event, _printed)


from typing import Annotated

from typing_extensions import TypedDict

from langgraph.graph.message import AnyMessage, add_messages

from datetime import datetime

from ylz_utils.cli.custom_service.car_rentals import book_car_rental, cancel_car_rental, search_car_rentals, update_car_rental
from ylz_utils.cli.custom_service.excursions import book_excursion, cancel_excursion, search_trip_recommendations, update_excursion
from ylz_utils.cli.custom_service.flights import cancel_ticket, fetch_user_flight_information, search_flights, update_ticket_to_new_flight
from ylz_utils.cli.custom_service.hotels import book_hotel, cancel_hotel, search_hotels, update_hotel
from ylz_utils.cli.custom_service.policies import lookup_policy
from ylz_utils.cli.custom_service.utils import create_tool_node_with_fallback

class State(TypedDict):
    messages: Annotated[list[AnyMessage], add_messages]

from langchain_openai import ChatOpenAI
from langchain_community.tools.tavily_search import TavilySearchResults
from langchain_core.prompts import ChatPromptTemplate
from langchain_core.runnables import Runnable, RunnableConfig

from ylz_utils.langchain import LangchainLib

class Assistant:
    def __init__(self, runnable: Runnable):
        self.runnable = runnable

    def __call__(self, state: State, config: RunnableConfig):
        while True:
            configuration = config.get("configurable", {})
            passenger_id = configuration.get("passenger_id", None)
            state = {**state, "user_info": passenger_id}
            result = self.runnable.invoke(state)
            # If the LLM happens to return an empty response, we will re-prompt it
            # for an actual response.
            if not result.tool_calls and (
                not result.content
                or isinstance(result.content, list)
                and not result.content[0].get("text")
            ):
                messages = state["messages"] + [("user", "Respond with a real output.")]
                state = {**state, "messages": messages}
            else:
                break
        return {"messages": result}


# Haiku is faster and cheaper, but less accurate
# llm = ChatAnthropic(model="claude-3-haiku-20240307")
langchainLib = LangchainLib()
llm = langchainLib.get_llm(temperature=1)
# You could swap LLMs, though you will likely want to update the prompts when
# doing so!
# from langchain_openai import ChatOpenAI

# llm = ChatOpenAI(model="gpt-4-turbo-preview")

primary_assistant_prompt = ChatPromptTemplate.from_messages(
    [
        (
            "system",
            "You are a helpful customer support assistant for Swiss Airlines. "
            " Use the provided tools to search for flights, company policies, and other information to assist the user's queries. "
            " When searching, be persistent. Expand your query bounds if the first search returns no results. "
            " If a search comes up empty, expand your search before giving up."
            "\n\nCurrent user:\n\n{user_info}\n"
            "\nCurrent time: {time}.",
        ),
        ("placeholder", "{messages}"),
    ]
).partial(time=datetime.now())

tavily_tool = langchainLib.toolLib.web_search.get_tool(rows=1)
part_1_tools = [
    #TavilySearchResults(max_results=1),
    tavily_tool,
    fetch_user_flight_information,
    search_flights,
    lookup_policy,
    update_ticket_to_new_flight,
    cancel_ticket,
    search_car_rentals,
    book_car_rental,
    update_car_rental,
    cancel_car_rental,
    search_hotels,
    book_hotel,
    update_hotel,
    cancel_hotel,
    search_trip_recommendations,
    book_excursion,
    update_excursion,
    cancel_excursion,
]
part_1_assistant_runnable = primary_assistant_prompt | llm.bind_tools(part_1_tools)

from langgraph.checkpoint.sqlite import SqliteSaver
from langgraph.graph import END, StateGraph, START
from langgraph.prebuilt import tools_condition

builder = StateGraph(State)


# Define nodes: these do the work
builder.add_node("assistant", Assistant(part_1_assistant_runnable))
builder.add_node("tools", create_tool_node_with_fallback(part_1_tools))
# Define edges: these determine how the control flow moves
builder.add_edge(START, "assistant")
builder.add_conditional_edges(
    "assistant",
    tools_condition,
)
builder.add_edge("tools", "assistant")

# The checkpointer lets the graph persist its state
# this is a complete memory for the entire graph.
memory = SqliteSaver.from_conn_string(":memory:")
part_1_graph = builder.compile(checkpointer=memory)

__all__=["part_1_graph"]


from langchain_community.document_loaders import UnstructuredFileLoader
from ylz_utils import LangchainLib

def start_rag(langchainLib:LangchainLib,args):
    embedding_key = args.embedding_key
    embedding_model = args.embedding_model
    rag_dbname = args.rag_dbname or "rag.faiss"
    url = args.url
    pptx = args.pptx
    docx = args.docx
    pdf = args.pdf
    glob = args.glob
    message = args.message
    
    
    if (not url and not pptx and not docx and not pdf and not glob) and  (not message):
        print(f"1、指定url/pptx/docx:系统将文档下载切片后向量化到{rag_dbname}数据库\n2、指定message:系统将从{rag_dbname}数据库中搜索相关的两条记录。\n您需要至少指定url和message中的一个参数.")
        return

    if embedding_key or embedding_model:
        embedding = langchainLib.get_embedding(embedding_key,embedding_model)
    else:
        embedding = None

    if (url or pptx or docx or pdf or glob):
        docs = start_loader(langchainLib,args)
        print("#"*60)
        if not docs:
            return
        if url and rag_dbname:
            ##### create vectorestore
            # url = "https://python.langchain.com/v0.2/docs/concepts/#tools"
            # faiss_dbname = "langchain_docs.faiss"
            print("result:",[{"doc_len":len(doc['doc'].page_content),"doc_blocks":len(doc['blocks'])} for doc in docs])
            for doc in docs:
                blocks = doc['blocks']
                if rag_dbname.startswith("es:///"):
                    _,index_name  =langchainLib.vectorstoreLib.esLib.init_client(connect_string=rag_dbname)
                    vectorstore = langchainLib.vectorstoreLib.esLib.get_store(index_name,embedding)
                    ids = langchainLib.vectorstoreLib.esLib.create_from_docs(vectorstore,blocks)                    
                else:
                    vectorstore,ids = langchainLib.vectorstoreLib.faissLib.create_from_docs(blocks,embedding)
                    langchainLib.vectorstoreLib.faissLib.save(rag_dbname,vectorstore)
                print("ids:",ids)
        else:
                if rag_dbname.startswith("es:///"):
                    _,index_name = langchainLib.vectorstoreLib.esLib.init_client(connect_string=rag_dbname)
                    vectorstore = langchainLib.vectorstoreLib.esLib.get_store(index_name,embedding)
                    ids = langchainLib.vectorstoreLib.esLib.create_from_docs(vectorstore,docs)   
                else:
                    vectorstore,ids = langchainLib.vectorstoreLib.faissLib.create_from_docs(docs,embedding)
                    langchainLib.vectorstoreLib.faissLib.save(rag_dbname,vectorstore)
                print("ids:",ids)
    if message and rag_dbname:   
        # docs = [Document("I am a student"),Document("who to go to china"),Document("this is a table")]
        # vectorestore = langchainLib.vectorstoreLib.faissLib.create_from_docs(docs)
        # langchainLib.vectorstoreLib.faissLib.save("test.faiss",vectorestore)
        
        vectorstore = langchainLib.vectorstoreLib.faissLib.load(rag_dbname,embedding)
        print("v--->",langchainLib.vectorstoreLib.faissLib.search(message,vectorstore,k=2))
    
    ###### have bug when poetry add sentence_transformers   
    #v1 = langchainLib.get_huggingface_embedding()
    #print("huggingface BGE:",v1)

    ###### tet google embdeeding
    # embed = langchainLib.get_embedding("EMBEDDING.GEMINI")
    # docs = [Document("I am a student"),Document("who to go to china"),Document("this is a table")]
    # vectorestore = langchainLib.vectorstoreLib.faissLib.create_from_docs(docs,embedding=embed)
    # langchainLib.vectorstoreLib.faissLib.save("test.faiss",vectorestore,index_name="gemini")

def start_loader(langchainLib:LangchainLib,args):
    url = args.url
    depth = args.depth
    docx_file = args.docx
    pptx_file = args.pptx
    pdf_file = args.pdf
    glob = args.glob
    chunk_size = args.size or 512
    only_one = list(filter(lambda x: x,[url,docx_file,pptx_file,pdf_file,glob]))
    if len(only_one) != 1:
        print(f"请指定url,docx,pptx,pdf,glob其中的一个")
        return 
    if url:
        result = langchainLib.load_url_and_split_markdown(url = url,max_depth = depth, chunk_size=chunk_size)
        print("result:",[{"doc_len":len(doc['doc'].page_content),"doc_blocks":len(doc['blocks']),"metadata":doc['metadata']} for doc in result])
    elif docx_file:
        result = langchainLib.documentLib.docx.load_and_split(docx_file,chunk_size=chunk_size)
        print(result)
    elif pptx_file:
        result = langchainLib.documentLib.pptx.load_and_split(pptx_file,chunk_size=chunk_size)
        print(result)
    elif pdf_file:
        result = langchainLib.documentLib.pdf.load_and_split(pdf_file,chunk_size=chunk_size)
        print(result)
    elif glob:
        loader_cls = UnstructuredFileLoader
        if glob.find(".docx")>=0:
            loader_cls = langchainLib.documentLib.docx.loader
        elif glob.find(".pptx")>=0:
            loader_cls = langchainLib.documentLib.pptx.loader
        elif glob.find(".pdf")>=0:
            loader_cls = langchainLib.documentLib.pdf.loader
        loader  = langchainLib.documentLib.dir.loader(".",glob=glob,show_progress=True,loader_cls=loader_cls)
        result  = loader.load_and_split(langchainLib.splitterLib.get_textsplitter(chunk_size=chunk_size,chunk_overlap=0))
        print(result)
    return result



import re
from langchain_core.documents import Document

import numpy as np
import openai
from langchain_core.tools import tool

import requests

from ylz_utils.langchain import LangchainLib

response = requests.get(
    "https://storage.googleapis.com/benchmarks-artifacts/travel-db/swiss_faq.md"
)
response.raise_for_status()
faq_text = response.text

#docs = [{"page_content": txt} for txt in re.split(r"(?=\n##)", faq_text)]
docs = [Document(txt) for txt in re.split(r"(?=\n##)", faq_text)]


class VectorStoreRetriever:
    def __init__(self, docs: list, vectors: list, oai_client):
        self._arr = np.array(vectors)
        self._docs = docs
        self._client = oai_client

    @classmethod
    def from_docs(cls, docs, oai_client):
        embeddings = oai_client.embeddings.create(
            model="text-embedding-3-small", input=[doc["page_content"] for doc in docs]
        )
        vectors = [emb.embedding for emb in embeddings.data]
        return cls(docs, vectors, oai_client)

    def query(self, query: str, k: int = 5) -> list[dict]:
        embed = self._client.embeddings.create(
            model="text-embedding-3-small", input=[query]
        )
        # "@" is just a matrix multiplication in python
        scores = np.array(embed.data[0].embedding) @ self._arr.T
        top_k_idx = np.argpartition(scores, -k)[-k:]
        top_k_idx_sorted = top_k_idx[np.argsort(-scores[top_k_idx])]
        return [
            {**self._docs[idx], "similarity": scores[idx]} for idx in top_k_idx_sorted
        ]


#retriever = VectorStoreRetriever.from_docs(docs, openai.Client())
langchainLib = LangchainLib()
embedding = langchainLib.embeddingLib.get_embedding(key="EMBEDDING.OLLAMA")
retriever = langchainLib.vectorstoreLib.faiss.create_from_docs(docs,embedding)


@tool
def lookup_policy(query: str) -> str:
    """Consult the company policies to check whether certain options are permitted.
    Use this before making any flight changes performing other 'write' events."""
    #docs = retriever.query(query, k=2)
    docs = langchainLib.vectorstoreLib.faiss.search(query,retriever,k=2)
    return "\n\n".join([doc["page_content"] for doc in docs])


from datetime import datetime
from elasticsearch_dsl import Index as dsl_Index
from elasticsearch_dsl import connections, Field,Document, Date, Nested, Boolean, \
    analyzer, InnerDoc, Completion, Keyword, Text,Integer,Long,Double,Float,\
    DateRange,IntegerRange,FloatRange,IpRange,Ip,Range
from elasticsearch_dsl import Search,Q,FacetedSearch,MultiSearch
from elasticsearch_dsl.query import MultiMatch
from typing import Optional,Union,Literal

from ylz_utils.config import Config

class ESLib():
    using = None
    analyzer='ik_max_word'
    search_analyzer='ik_smart'
    indexes = {}
    def __init__(self,hosts:str=None,
                 es_user:str=None ,
                 es_password:str=None ,
                 using:str = "es",
                 verify_certs=False,
                 ssl_show_warn=False,
                 analyzer='ik_max_word',search_analyzer='ik_smart'):
        self.config = Config()
        self.using = using
        self.analyzer = analyzer
        self.search_analyzer = search_analyzer
        connections.create_connection(
            hosts=hosts or self.config.get("ES.HOST"),
            basic_auth=(es_user or self.config.get("ES.USER"),es_password or self.config.get("ES.PASSWORD")),
            alias=using,
            verify_certs=verify_certs,
            ssl_show_warn=ssl_show_warn
        )
        self.Q = Q
        self.Search = Search
        self.MultiSearch = MultiSearch
        self.FacetedSearch = FacetedSearch
        self.MultiMatch = MultiMatch
    
    def register_class(self,cls:Document):
        setattr(self, cls.__name__, cls)
        self.indexes[cls.Index.name]=cls 
        cls.init(using=self.using) 
        return cls
    def _parse_fields(self,raw_fields):
        wrapped_fields : dict[str,Field]= {}
        for key in raw_fields:
            field_type = raw_fields[key].get("type")
            multi = raw_fields[key].get("multi")
            required = raw_fields[key].get("requried")
            if field_type=='int':
                wrapped_fields[key] = Integer(multi=multi,required=required)
            elif field_type=='long':
                wrapped_fields[key] = Long(multi=multi,required=required)
            elif field_type=='float':
                wrapped_fields[key] = Float(multi=multi,required=required)
            elif field_type=='double':
                wrapped_fields[key] = Double(multi=multi,required=required)
            elif field_type=='boolean':
                wrapped_fields[key] = Boolean(multi=multi,required=required)
            elif field_type=='ktext':
                wrapped_fields[key] = Text(analyzer=self.analyzer,search_analyzer=self.search_analyzer,fields={'keyword':Keyword()},multi=multi,required=required)
            elif field_type=='text':
                wrapped_fields[key] = Text(analyzer=self.analyzer,search_analyzer=self.search_analyzer,multi=multi,required=required)
            elif field_type=='keyword':
                wrapped_fields[key] = Keyword(multi=multi,required=required)
            elif field_type=='datetime':
                wrapped_fields[key] = Date(multi=multi,required=required)
            elif field_type=='date':
                wrapped_fields[key] = Date(format="yyyy-MM-dd", multi=multi,required=required)
            elif field_type=='ip':
                wrapped_fields[key] = Ip(multi=multi,required=required)
            elif field_type=='date_range':
                wrapped_fields[key] = DateRange(multi=multi,required=required)
            elif field_type=='ip_range':
                wrapped_fields[key] = IpRange(multi=multi,required=required)
            elif field_type=='int_range':
                wrapped_fields[key] = IntegerRange(multi=multi,required=required)
            elif field_type=='float_range':
                wrapped_fields[key] = FloatRange(multi=multi,required=required)
            elif field_type=='innerdoc':
                innerdoc = raw_fields[key]["innerdoc"]
                innerdoc_name = innerdoc["name"]
                innerdoc_raw_fields = innerdoc["fields"]
                innerdoc_wrapped_fields = self._parse_fields(innerdoc_raw_fields)
                innerdoc_cls = type(innerdoc_name, (InnerDoc,), innerdoc_wrapped_fields)
                setattr(self,innerdoc_cls.__name__,innerdoc_cls) 
                #wrapped_fields[key] = innerdoc_cls() ????
            else:
                raise Exception(f"ESLib不支持的数据类型:{field_type}") 
        return wrapped_fields               

    def register(self, class_name, index_name, shards:int=None,replicas:int=None,
                 fields:dict[str,dict]={}):
        """动态创建Document类。

        Args:
            class_name (str): 类名。
            index_name (str): index的名称
            shards (int): shards数量
            replicas (int): replicas数量
            field (key:{"type":Literal['int','long','float',\
                                         'double','boolean','ktext',\
                                         'text','keyword','datetime','date','ip' \
                                         'int_range','float_range','date_range','ip_range'
                                         ],
                        "required":False,
                        "multi":False}): 字段字典。 默认为空字典。

        Returns:
            type: 新创建Document类。
        """
        base_classes=(Document,)
        wrapped_fields = self._parse_fields(fields)
        # 创建 Index 类
        settings={}
        if shards:
            settings["number_of_shards"] = shards
        if replicas:
            settings["number_of_replicas"] = replicas
        Index = type('Index', tuple(), {'name': index_name, 'settings': settings})
        wrapped_fields['Index'] = Index
        cls = type(class_name, base_classes, wrapped_fields)        
        setattr(self, cls.__name__, cls)
        self.indexes[index_name] = cls
        cls.init(using=self.using)
        return cls
    def get_matchQ(self,column_name,value):
        return Q({"match":{column_name:{"query":value,"analyzer":"ik_smart"}}})
    def get_termQ(self,column_name,value):
        return Q({"term":{column_name + ".keyword":value}})
    def get_multimatchQ(self,column_names,value):
        return MultiMatch(query=value,fields=column_names,analyzer=self.search_analyzer)
    def get_search(self,index_name):
        return Search(using = self.using,index=index_name)
    def index_search(self,index_name,q:Q=None,s:Search|None=None,start=0,end=9):
        if not s:
            s = Search(using = self.using,index=index_name)
        if q:
            s=s.query(q)
        
        res = s[start:end].execute()
        cls = self.indexes[index_name]
        results = [cls(**data.to_dict()) for data in res]
        #size = res.hits.total.value
        return results
    def index_get(self,index_name,id:str,missing:Literal["none","raise","skip"]="none"):
        cls = self.indexes[index_name]
        doc = cls.get(using=self.using,id=id,missing=missing)
        return doc
    def index_mget(self,index_name,ids:list[str],missing:Literal["none","raise","skip"]="none"):
        cls = self.indexes[index_name]
        doc = cls.mget(using=self.using,docs=ids,missing=missing)
        return doc
    def index_exists(self,index_name) -> bool:
        index = dsl_Index(index_name)
        return index.exists_alias(using=self.using,name=index_name)
    # def index_create(self,index_name) -> bool:
    #     cls = self.indexes[index_name]
    #     return cls.save(using=self.using)
    def index_delete(self,index_name) -> bool:
        index = dsl_Index(index_name)
        return index.delete(using=self.using)
    def doc_dict(self,doc):
        return doc.to_dict()
    def doc_save(self,doc):
        return doc.save(using=self.using)
    def doc_update(self,doc,**kwargs):
        return doc.update(using=self.using,**kwargs)
    def doc_delete(self,doc):
        return doc.delete(using=self.using)
        
if __name__ == '__main__':
    Config.init('ylz_utils')
    esLib = ESLib(using='es')
    class Product(Document):
        id = Integer()
        name = Text(analyzer='ik_max_word',search_analyzer='ik_smart',fields={'keyword':Keyword()})
        summary = Text(analyzer='ik_max_word',search_analyzer='ik_smart')
        solution_class = Text(analyzer='ik_max_word',search_analyzer='ik_smart')
        tags = Keyword()
        price = Float()
        published_from = Date()
        class Index():
            name = 'product'
            settings = {
                "number_of_shards": 1,
                "number_of_replicas": 0,
            }
    esLib.register_class(Product)
    # data = [ 
    #     {"id":1,"name":'碗',"summary":'镶边且印有小狐狸图案',"solution_class":'home',"tags":['little'],"price":20.2},
    #     {"id":2,"name":'筷子',"summary":'很多，黑色的高级筷子',"solution_class":'home',"tags":['black','multi'],"price":5.8},
    #     {"id":3,"name":'钥匙',"summary":'专用的多功能开锁神器',"solution_class":'tool',"tags":['useful'],"price":3.00},
    #     {"id":4,"name":'手机',"summary":'各种型号的大屏幕，大内存手机',"solution_class":'office',"tags":['good','multi'],"price":2888},
    #     {"id":5,"name":'打火机',"summary":'积压仓库，准备清仓',"solution_class":'tool',"tags":['cheap'],"price":0.1},
    #     {"id":6,"name":'鼠标',"summary":'白色无线鼠标，不需要更换电池，typeC充电',"solution_class":'office',"tags":['wireless','white'],"price":109},
    #     {"id":7,"name":'苹果',"summary":'20斤大苹果，包甜',"solution_class":'food',"tags":['red','good'],"price":2.2},
    #     {"id":8,"name":'电脑',"summary":'苹果imac M1，M2，M3系列',"solution_class":'office',"tags":['useful'],"price":4888},
    #     {"id":9,"name":'桌子',"summary":'减价销售，售完为止',"solution_class":'office',"tags":['cheap'],"price":260},
    #     {"id":10,"name":'足球',"summary":'官方认证，品质保障',"solution_class":'sport',"tags":['useful'],"price":60},
    #     {"id":11,"name":'面巾纸',"summary":'丝滑柔软',"solution_class":'home',"tags":['little','multi'],"price":40.5},
    #     {"id":12,"name":'水壶',"summary":'大容量，装水充足，自动报警，智能烧水',"solution_class":'home',"tags":['useful','cheap'],"price":99},
    #     {"id":13,"name":'自行车',"summary":'风华牌28寸，26寸都有',"solution_class":'home',"tags":['useful'],"price":120},
    #     {"id":14,"name":'手电筒',"summary":'4节电池，多功能',"solution_class":'tool',"tags":['useful'],"price":15},
    #     {"id":15,"name":'沙发',"summary":'墨绿色，豪华舒适',"solution_class":'home',"tags":['big','cheap'],"price":1500},
    #     {"id":16,"name":'香烟',"summary":'23条，吸烟有害健康',"solution_class":'food',"tags":['red'],"price":17.8},
    #     ]
    # products = []
    # for item in data:
    #     product = Product(
    #         meta={'id':item['id']},id=item['id'],name=item['name'],
    #         summary=item['summary'],solution_class=item['solution_class'],tags=item['tags'],price=item['price'])
    #     products.append(product)
    #     esLib.doc_save(product)
    products = esLib.index_search("product")
    print("class search:",[esLib.doc_dict(product) for product in products])    
    # product.save(using="es")
    #product = esLib.Product.get(id=42,using='es')
    #print(product)
    ProductNew = esLib.register("Product","product_new",fields={
        "id": {"type":"int","required":True},
        "name":{"type":"ktext"},
        "summary":{"type":"text"},
        "tags":{"type":"keyword","multi":True},
        "published_from":{"type":"date",},
        "info":{"type":"innerdoc",
                "innerdoc":{"name":"Info",
                            "fields":{
                                "provider":{"type":"text"},
                                "price":{"type":"float"}
        }}}
    })
    # print("index_exists:",esLib.index_exists("product_new"))
    # docs = esLib.index_search("product_new")
    # product1 = ProductNew(meta={'id':44},id=1044,name='soup',summary='bowl with a little fox',solution_class='home',tags=['Linabell'])
    # product1.info = esLib.Info(provider="abc",price=20.2)
    # print("*"*20,product1.to_dict())
    # product2 = ProductNew(meta={'id':34},id=1034,name='mouse',summary='a white mouse',solution_class='home',tags=['white','wireless'])
    # product2.info = esLib.Info(provider="xyz",price=30.4)
    # print("to_dict:",esLib.doc_dict(product1))
    # print("insert:",esLib.doc_save(product1))
    # print("insert:",esLib.doc_save(product2))
    # print(esLib.index_mget("product_new",ids=[44,34]))
    # print("update:",esLib.doc_update(product1,summary="hello world"))
    # print(esLib.index_mget("product_new",ids=[44,34]))
    # #print("delete:",esLib.doc_delete(product1))
    # print(esLib.index_mget("product_new",ids=[44,34]))
    # #print("index_delete",esLib.index_delete("product_new"))
    
    # # for doc in docs:
    # #     esLib.delete(doc)
    # #product.save(using="es")
    # #esLib.save("product_new")  
    # # product = Product(meta={'id':44},id=1044,name='soup',summary='bowl with a little fox',solution_class='home',tags='Linabell')
    # # product.publish_from = datetime.now()
    # # product.save(using="es")
    
    # # q = esLib.termQ("tags","Linabell")
    # # res,size = esLib.search("product",q) 
    # # if res:
    # #     print(res[0].name,size)
    # # else:
    # #     print("no result")
    
    # # s=esLib.get_search("product")
    # # s.query("term",solution_class="abc")
    # # res = s.execute()
    # # print([(hit.meta.score , hit.name) for hit in res.hits])
    
    # results = esLib.index_search("product")
    # print([result.to_dict() for result in results])

    s = esLib.get_search("product")
    s.query("term",name="手机")
    results = esLib.index_search("product",s=s)
    print([result.to_dict() for result in results])


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib

from langchain.memory import ConversationBufferMemory
from langchain_elasticsearch import ElasticsearchChatMessageHistory
from langchain_community.chat_message_histories import SQLChatMessageHistory,ChatMessageHistory,ElasticsearchChatMessageHistory
from langchain_elasticsearch.client import create_elasticsearch_client

from ylz_utils.config import Config 

class MemoryLib():
    def __init__(self,langchainLib:LangchainLib):
        self.langchainLib = langchainLib
        self.config = Config()
    def get_memory(self,dbname:str,user_id:str,conversation_id:str):
        if dbname:
            if dbname.startswith("es://"):
                # es://username:password@127.0.0.1:9200/index_name 
                # es:///index_name
                # sqlite:///db_name
                # db_name
                # :memory:
                if dbname.startswith("es:///"):
                    es_url = self.config.get("ES.HOST")
                    es_username = self.config.get("ES.USER")
                    es_password = self.config.get("ES.PASSWORD")
                    index_name = dbname.split(':///')[1]
                else:
                    es_url = f"https://{dbname.split('@')[1].split('/')[0]}"
                    es_username = dbname.split('@')[0].split('://')[1].split(":")[0]
                    es_password = dbname.split('@')[0].split('://')[1].split(":")[1]
                    index_name = dbname.split('@')[1].split('/')[1]
                session_id = f"{user_id}--{conversation_id}"
                #print(es_url,es_username,es_password,index_name,session_id)
                es_connection = create_elasticsearch_client(
                                url=es_url,
                                username=es_username,
                                password=es_password,
                                params = {"verify_certs":False,"ssl_show_warn":False},
                            )
                return  ElasticsearchChatMessageHistory(
                            es_connection=es_connection,
                            index=index_name,
                            session_id=session_id
                        )
            elif dbname.startswith("sqlite:///"):
                return SQLChatMessageHistory(f"{user_id}--{conversation_id}", f"{dbname}")
            elif dbname==":memory:":
                return SQLChatMessageHistory(f"{user_id}--{conversation_id}", ":memory:")
            else:
                return SQLChatMessageHistory(f"{user_id}--{conversation_id}", f"sqlite:///{dbname}")
        raise Exception("you must give the chat_dbname . please use set_dbname first!")



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib
    
import re

from langchain.docstore.document import Document
from langchain_text_splitters import RecursiveCharacterTextSplitter
from langchain_text_splitters import MarkdownHeaderTextSplitter

class SplitterLib():

    def get_textsplitter(self,chunk_size=1000,chunk_overlap=10):
        text_splitter = RecursiveCharacterTextSplitter(
                chunk_size=chunk_size, chunk_overlap=chunk_overlap
            )
        return text_splitter
    
    def split_markdown_docs(self,text,chunk_size=1000,chunk_overlap=0):
            splited_result = self.split_text_with_protected_blocks(text,chunk_size=chunk_size,chunk_overlap=chunk_overlap)
            # Split
            splited_docs = list(map(lambda item:Document(page_content=item),splited_result))
            return splited_docs
    def extract_blocks(self,text, pattern):
        """通用函数来提取符合模式的块"""
        blocks = pattern.findall(text)
        return blocks
    def replace_blocks_with_placeholders(self,text, blocks, block_type):
        """使用占位符替换块"""
        for i, block in enumerate(blocks):
            text = text.replace(block, f'{{{block_type}_{i}}}')
        return text
    def restore_blocks(self,text, blocks, block_type):
        """将占位符替换回块"""
        for i, block in enumerate(blocks):
            text = text.replace(f'{{{block_type}_{i}}}', block)
        return text
    def split_text(self,text,chunk_size=1000,chunk_overlap=0):
        """你的拆分逻辑，例如按段落拆分"""
        #return text.split('\n\n')
        text_splitter = RecursiveCharacterTextSplitter(
                chunk_size=chunk_size, chunk_overlap=chunk_overlap
            )
        # Split
        splited_docs = text_splitter.split_documents([Document(page_content=text)])

        return map(lambda item:item.page_content,splited_docs)
    def split_text_with_protected_blocks(self,text,chunk_size,chunk_overlap):
        # 定义匹配Markdown表格的正则表达式
        #table_pattern = re.compile(
        # r'''
        # (                           # 捕获组
        #     ^\|.*\|[\r\n]+          # 表头行
        #     (?:\|[-\s:]*\|[\r\n]*)  # 分隔行
        #     (?:\|.*\|[\r\n]*)+      # 数据行
        # )
        # ''', 
        # re.MULTILINE | re.VERBOSE
        # )
        table_pattern = re.compile(
        r'''
        (                           # 捕获组
            ^\|.*\|.*$              # 表头行
            (?:\r?\n\|.*\|.*$)+     # 后续行
        )
        ''', 
        re.MULTILINE | re.VERBOSE
        )
        # 定义匹配脚本代码块的正则表达式
        script_pattern = re.compile(r'((?: {4}.+\n)+)', re.MULTILINE)
        #script_pattern = re.compile(r"^(\t|(?:\n))*(?:```)(.*?)```", re.MULTILINE)

        # 提取表格和脚本块
        tables = self.extract_blocks(text, table_pattern)
        scripts = self.extract_blocks(text, script_pattern)
        
        # 用占位符替换表格和脚本块
        text_with_placeholders = self.replace_blocks_with_placeholders(text, tables, 'TABLE')
        text_with_placeholders = self.replace_blocks_with_placeholders(text_with_placeholders, scripts, 'SCRIPT')
        
        #FileLib.writeFile("current.md",text_with_placeholders)
        # 拆分文本
        split_parts = self.split_text(text_with_placeholders,chunk_size=chunk_size,chunk_overlap=chunk_overlap)
        
        # 恢复表格和脚本块
        restored_parts = [self.restore_blocks(part, tables, 'TABLE') for part in split_parts]
        restored_parts = [self.restore_blocks(part, scripts, 'SCRIPT') for part in restored_parts]
        
        return restored_parts



import os
import shutil
import sqlite3

import pandas as pd
import requests

db_url = "https://storage.googleapis.com/benchmarks-artifacts/travel-db/travel2.sqlite"
local_file = "travel2.sqlite"
# The backup lets us restart for each tutorial section
backup_file = "travel2.backup.sqlite"
overwrite = False
if overwrite or not os.path.exists(local_file):
    response = requests.get(db_url)
    response.raise_for_status()  # Ensure the request was successful
    with open(local_file, "wb") as f:
        f.write(response.content)
    # Backup - we will use this to "reset" our DB in each section
    shutil.copy(local_file, backup_file)
# Convert the flights to present time for our tutorial
conn = sqlite3.connect(local_file)
cursor = conn.cursor()

tables = pd.read_sql(
    "SELECT name FROM sqlite_master WHERE type='table';", conn
).name.tolist()
tdf = {}
for t in tables:
    tdf[t] = pd.read_sql(f"SELECT * from {t}", conn)

example_time = pd.to_datetime(
    tdf["flights"]["actual_departure"].replace("\\N", pd.NaT)
).max()
current_time = pd.to_datetime("now").tz_localize(example_time.tz)
time_diff = current_time - example_time

tdf["bookings"]["book_date"] = (
    pd.to_datetime(tdf["bookings"]["book_date"].replace("\\N", pd.NaT), utc=True)
    + time_diff
)

datetime_columns = [
    "scheduled_departure",
    "scheduled_arrival",
    "actual_departure",
    "actual_arrival",
]
for column in datetime_columns:
    tdf["flights"][column] = (
        pd.to_datetime(tdf["flights"][column].replace("\\N", pd.NaT)) + time_diff
    )

for table_name, df in tdf.items():
    df.to_sql(table_name, conn, if_exists="replace", index=False)
del df
del tdf
conn.commit()
conn.close()

db = local_file  # We'll be using this local file as our DB in this tutorial

__all__ = ["db","backup_file"]


import sqlite3
from datetime import date, datetime
from typing import Optional

import pytz
from langchain_core.runnables import ensure_config

from langchain_core.tools import tool

from ylz_utils.cli.custom_service.db_init import db

@tool
def fetch_user_flight_information() -> list[dict]:
    """Fetch all tickets for the user along with corresponding flight information and seat assignments.

    Returns:
        A list of dictionaries where each dictionary contains the ticket details,
        associated flight details, and the seat assignments for each ticket belonging to the user.
    """
    config = ensure_config()  # Fetch from the context
    configuration = config.get("configurable", {})
    passenger_id = configuration.get("passenger_id", None)
    if not passenger_id:
        raise ValueError("No passenger ID configured.")

    conn = sqlite3.connect(db)
    cursor = conn.cursor()

    query = """
    SELECT 
        t.ticket_no, t.book_ref,
        f.flight_id, f.flight_no, f.departure_airport, f.arrival_airport, f.scheduled_departure, f.scheduled_arrival,
        bp.seat_no, tf.fare_conditions
    FROM 
        tickets t
        JOIN ticket_flights tf ON t.ticket_no = tf.ticket_no
        JOIN flights f ON tf.flight_id = f.flight_id
        JOIN boarding_passes bp ON bp.ticket_no = t.ticket_no AND bp.flight_id = f.flight_id
    WHERE 
        t.passenger_id = ?
    """
    cursor.execute(query, (passenger_id,))
    rows = cursor.fetchall()
    column_names = [column[0] for column in cursor.description]
    results = [dict(zip(column_names, row)) for row in rows]

    cursor.close()
    conn.close()

    return results


@tool
def search_flights(
    departure_airport: Optional[str] = None,
    arrival_airport: Optional[str] = None,
    start_time: Optional[date | datetime] = None,
    end_time: Optional[date | datetime] = None,
    limit: int = 20,
) -> list[dict]:
    """Search for flights based on departure airport, arrival airport, and departure time range."""
    conn = sqlite3.connect(db)
    cursor = conn.cursor()

    query = "SELECT * FROM flights WHERE 1 = 1"
    params = []

    if departure_airport:
        query += " AND departure_airport = ?"
        params.append(departure_airport)

    if arrival_airport:
        query += " AND arrival_airport = ?"
        params.append(arrival_airport)

    if start_time:
        query += " AND scheduled_departure >= ?"
        params.append(start_time)

    if end_time:
        query += " AND scheduled_departure <= ?"
        params.append(end_time)
    query += " LIMIT ?"
    params.append(limit)
    cursor.execute(query, params)
    rows = cursor.fetchall()
    column_names = [column[0] for column in cursor.description]
    results = [dict(zip(column_names, row)) for row in rows]

    cursor.close()
    conn.close()

    return results


@tool
def update_ticket_to_new_flight(ticket_no: str, new_flight_id: int) -> str:
    """Update the user's ticket to a new valid flight."""
    config = ensure_config()
    configuration = config.get("configurable", {})
    passenger_id = configuration.get("passenger_id", None)
    if not passenger_id:
        raise ValueError("No passenger ID configured.")

    conn = sqlite3.connect(db)
    cursor = conn.cursor()

    cursor.execute(
        "SELECT departure_airport, arrival_airport, scheduled_departure FROM flights WHERE flight_id = ?",
        (new_flight_id,),
    )
    new_flight = cursor.fetchone()
    if not new_flight:
        cursor.close()
        conn.close()
        return "Invalid new flight ID provided."
    column_names = [column[0] for column in cursor.description]
    new_flight_dict = dict(zip(column_names, new_flight))
    timezone = pytz.timezone("Etc/GMT-3")
    current_time = datetime.now(tz=timezone)
    departure_time = datetime.strptime(
        new_flight_dict["scheduled_departure"], "%Y-%m-%d %H:%M:%S.%f%z"
    )
    time_until = (departure_time - current_time).total_seconds()
    if time_until < (3 * 3600):
        return f"Not permitted to reschedule to a flight that is less than 3 hours from the current time. Selected flight is at {departure_time}."

    cursor.execute(
        "SELECT flight_id FROM ticket_flights WHERE ticket_no = ?", (ticket_no,)
    )
    current_flight = cursor.fetchone()
    if not current_flight:
        cursor.close()
        conn.close()
        return "No existing ticket found for the given ticket number."

    # Check the signed-in user actually has this ticket
    cursor.execute(
        "SELECT * FROM tickets WHERE ticket_no = ? AND passenger_id = ?",
        (ticket_no, passenger_id),
    )
    current_ticket = cursor.fetchone()
    if not current_ticket:
        cursor.close()
        conn.close()
        return f"Current signed-in passenger with ID {passenger_id} not the owner of ticket {ticket_no}"

    # In a real application, you'd likely add additional checks here to enforce business logic,
    # like "does the new departure airport match the current ticket", etc.
    # While it's best to try to be *proactive* in 'type-hinting' policies to the LLM
    # it's inevitably going to get things wrong, so you **also** need to ensure your
    # API enforces valid behavior
    cursor.execute(
        "UPDATE ticket_flights SET flight_id = ? WHERE ticket_no = ?",
        (new_flight_id, ticket_no),
    )
    conn.commit()

    cursor.close()
    conn.close()
    return "Ticket successfully updated to new flight."


@tool
def cancel_ticket(ticket_no: str) -> str:
    """Cancel the user's ticket and remove it from the database."""
    config = ensure_config()
    configuration = config.get("configurable", {})
    passenger_id = configuration.get("passenger_id", None)
    if not passenger_id:
        raise ValueError("No passenger ID configured.")
    conn = sqlite3.connect(db)
    cursor = conn.cursor()

    cursor.execute(
        "SELECT flight_id FROM ticket_flights WHERE ticket_no = ?", (ticket_no,)
    )
    existing_ticket = cursor.fetchone()
    if not existing_ticket:
        cursor.close()
        conn.close()
        return "No existing ticket found for the given ticket number."

    # Check the signed-in user actually has this ticket
    cursor.execute(
        "SELECT flight_id FROM tickets WHERE ticket_no = ? AND passenger_id = ?",
        (ticket_no, passenger_id),
    )
    current_ticket = cursor.fetchone()
    if not current_ticket:
        cursor.close()
        conn.close()
        return f"Current signed-in passenger with ID {passenger_id} not the owner of ticket {ticket_no}"

    cursor.execute("DELETE FROM ticket_flights WHERE ticket_no = ?", (ticket_no,))
    conn.commit()

    cursor.close()
    conn.close()
    return "Ticket successfully cancelled."


from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib

from langchain_community.vectorstores import FAISS
from langchain_core.documents import Document
from tqdm import tqdm
from typing import List

class FaissLib():
    def __init__(self,langchainLib:LangchainLib):
        self.langchainLib = langchainLib

    def create_from_docs(self,docs,embedding=None) -> FAISS:
        if not embedding:
            embedding = self.langchainLib.get_embedding()
        #vectorstore = FAISS.from_documents(docs,embedding=embedding)
        vectorstore = FAISS.from_documents([Document(" ")],embedding) 
        #vectorstore.add_documents(docs,embedding=embedding)
        all_ids = []
        with tqdm(total= len(docs)) as pbar:
            for index,doc in enumerate(docs):
                if doc.page_content:
                    ids = vectorstore.add_documents([doc],embedding=embedding)
                    all_ids.extend(ids)
                pbar.update(1)
        return vectorstore,all_ids
    def create_from_texts(self,texts,embedding=None) -> FAISS:
        if not embedding:
            embedding = self.langchainLib.get_embedding()
        #vectorstore = FAISS.from_texts(textes, embedding=embedding)
        vectorstore = FAISS.from_texts([" "],embedding) 
        all_ids = []
        with tqdm(total= len(texts)) as pbar:
            for idx,text in enumerate(texts):
                if text:
                    ids = vectorstore.add_texts([text],embedding=embedding)
                    all_ids.extend(ids)
                pbar.update(1)
        return vectorstore,all_ids
    
    def delete(self,vectorstore: FAISS,ids: List[str] | None = None):
        return vectorstore.delete(ids) 
      
    def save(self,  db_file:str, vectorstore: FAISS,index_name:str = "index"):
        vectorstore.save_local(db_file,index_name)

    def load(self, db_file:str ,embedding=None, index_name:str = "index") -> FAISS:
        if not embedding:
            embedding = self.langchainLib.get_embedding()
        vectorestore = FAISS.load_local(db_file, embeddings=embedding, index_name=index_name, allow_dangerous_deserialization=True)
        return vectorestore
    
    def search(self,query,vectorstore: FAISS,k=10,filter=[]):
        return vectorstore.similarity_search(query,k=k,filter=filter)
    
    def search_with_score(self,query,vectorstore: FAISS,k=10,filter=[]):
        return vectorstore.similarity_search_with_score(query,k=k,filter=filter)




from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib

from langchain_community.document_loaders import DirectoryLoader

class DirLoader():
    def __init__(self,langchainLib:LangchainLib):
        self.langchainLib = langchainLib
        self.loader  = DirectoryLoader


from ylz_utils import Color, StringLib
from ylz_utils import FileLib
from ylz_utils import LangchainLib

from pydantic import BaseModel,Field
from typing import Literal,List
from pptx.util import Inches,Cm,Pt
from pptx.enum.shapes import MSO_SHAPE

from langchain.prompts import PromptTemplate,ChatPromptTemplate
from langchain.output_parsers import OutputFixingParser

from langchain_core.tools import tool
from langchain_core.runnables import RunnablePassthrough,RunnableLambda,RunnableParallel

from ylz_utils.cli.chat import start_chat
from ylz_utils.cli.graph import start_graph
from ylz_utils.cli.rag import start_loader,start_rag

def __agent(langchainLib:LangchainLib,args):
    llm_key = args.llm_key
    llm_model = args.llm_model
    message = args.message if args.message else "厦门今天天气怎么样?"
    llm = langchainLib.get_llm(llm_key,llm_model)
    prompt = langchainLib.get_prompt()
    tools = []
    tavily_tool = langchainLib.toolLib.web_search.get_tool()
    tools.append(tavily_tool)
    # langchain_docs_vectorestore = langchainLib.vectorstoreLib.faiss.load("langchain_docs.db")
    # langchain_docs_retriever = langchain_docs_vectorestore.as_retriever()
    # rag_tool = langchainLib.toolLib.rag_search.get_tool(langchain_docs_retriever,"langchain docs","langchain的文档")
    # tools.append(rag_tool)
    
    agent =  langchainLib.get_agent(llm,tools)
    res = agent.stream({"messages":[("user",message)]})
    for chunk in res:
        print(chunk , end="")
    print("\n\n",langchainLib.get_llm(full=True),"\n")



def __llm_test(langchainLib:LangchainLib,args):
    llm_key = args.llm_key
    llm_model = args.llm_model
    message = args.message
    llm = langchainLib.get_llm(llm_key,llm_model)
    res = llm.invoke(message)
    print("res:",res)
    print("llms:",[(item["type"],item["api_key"],item["used"]) for item in langchainLib.get_llm(full=True)])

def __outputParser_test(langchainLib:LangchainLib,args):
    message = args.message
    llm_key = args.llm_key
    class Food(BaseModel):
        name: str = Field(description="name of the food")
        place: str = Field(defualt="未指定",description="where to eat food?",examples=["家","公司","体育馆"])
        calories: float|None = Field(title="卡路里(千卡)",description="食物热量")
        amount: tuple[float,str] = Field(description="tuple of the value and unit of the food amount")
        time: str = Field(default = "未指定",description="the time of eating food,确保每个食物都有对应的时间，如果没有指定则为空字符串")
        health: Literal["健康","不健康","未知"]|None = Field(description="根据常识判断食物是否为健康食品",examples=["健康","不健康","未知"])
    class Foods(BaseModel):
        foods: List[Food]

    # Set up a parser + inject instructions into the prompt template.
    parser = langchainLib.get_outputParser(Foods)

    prompt1 = PromptTemplate(
        template="用中文分析句子的内容。如果没有指定食物热量则根据食物的名称和数量进行估计。判断食物是否为健康\n{format_instructions}\n{command}\n",
        input_variables=["command"],
        partial_variables={"format_instructions": parser.get_format_instructions()},
    )


    # And a query intended to prompt a language model to populate the data structure.
    fixed_parser = OutputFixingParser.from_llm(parser=parser, llm=langchainLib.get_llm(llm_key))
    # retry_parser = RetryOutputParser.from_llm(parser=parser, llm=langchainLib.get_llm())

    prompt2 = langchainLib.get_prompt(
            system_prompt = "用中文分析句子的内容。如果没有指定食物热量则根据食物的名称和数量进行估计。同时判断食物是否为健康",
            human_keys={"command":"输入的语句"},
            outputParser = parser,
            use_chat = False
        )

    #chain = prompt2 | langchainLib.get_llm() | fixed_parser
    
    chain = prompt2 | langchainLib.get_llm(llm_key) | parser
    
    # retry_chain = RunnableParallel(
    #     completion=chain, prompt_value=prompt1,
    #     ) | RunnableLambda(lambda x:retry_parser.parse_with_prompt(**x))

    #output = chain.invoke({"command": "我今天在家吃了3个麦当劳炸鸡和一个焦糖布丁，昨天8点在电影院吃了一盘20千卡的西红柿炒蛋"})
    output = chain.invoke({"command": message})
    #retry_parser.parse_with_prompt(output['completion'], output['prompt_value'])
    #retry_parser.parse_with_prompt(output['completion'],output['prompt_value'])
    print("\noutput parser:",output)

    print("llms:",[(item["type"],item["api_key"],item["used"]) for item in langchainLib.get_llm(full=True)])


def __runnalble_test(langchainLib:LangchainLib):
    runnable1 = RunnableParallel(
    passed=RunnablePassthrough(),
    modified=lambda x: x["num"] + 1,
    other = RunnableParallel(
        passed=RunnablePassthrough(),
        modified=lambda x: x["num"] * 2,
        )
    )
    #runnable1.invoke({"num":1})
    def f(x):
        return x['modified']*100

    runnable2 = runnable1 | RunnableParallel(
        {"origin": RunnablePassthrough(),
        "new1": lambda x: x["other"]["passed"],
        "new2": RunnableLambda(f)}
        )

    #for trunck in runnable2.stream({"num": 1}):
    #  print(trunck)
    res = runnable2.invoke({"num": 1})
    print("\nrunnalbe:",res)

def __prompt_test(langchainLib:LangchainLib,args):
    llm_key = args.llm_key
    llm_model = args.llm_model
    prompt = langchainLib.get_prompt(f"你对日历时间非常精通",human_keys={"context":"关联的上下文","question":"问题是"},use_chat = False)
    prompt.partial(context="我的名字叫小美")   # 这个为什么不起作用?
    print("!!!!",langchainLib.llmLib.default_llm_key,llm_key)
    llm = langchainLib.get_llm(llm_key,llm_model)
    print(llm)
    outputParser = langchainLib.get_outputParser()
    chain = prompt | llm | outputParser

    res = chain.invoke({"question":"我叫什么名字，今天礼拜几","context":"我是海涛"})
    print(res)
    res = chain.stream({"question":"用我的名字写一首诗歌","context":"我的名字叫小美"})
    for chunk in res:
        print(chunk,end="")
    
    prompt = PromptTemplate.from_template("用中文回答：{topic} 的{what}是多少")
    llm = langchainLib.get_llm("LLM.GROQ")
    print(llm)
    chain = prompt | llm | outputParser
    promise = chain.batch([{"topic":"中国","what":"人口"},{"topic":"美国","what":"国土面积"},{"topic":"新加坡","what":"大学"}])
    print(promise)
    print("llms:",[(item["type"],item["api_key"],item["used"]) for item in langchainLib.get_llm(full=True)])


def __tools_test(langchainLib:LangchainLib,args):
    tool = langchainLib.toolLib.python_repl.get_tool()
    res = tool.invoke(
"""
print(3+2)
""")
    print("python repl=",res)
    tool = langchainLib.toolLib.wolfram_alpha.get_tool()
    res = tool.run("what is the square root of 25?")
    print(res)
    return 
    # tool: TavilySearchResults = langchainLib.get_search_tool("TAVILY")
    # prompt = langchainLib.get_prompt(is_chat=False,human_keys={"context":"关联的上下文是:","question":"问题是:"})
    # res = tool.invoke("易联众现在股价是多少？")
    # print(res)
    # llm = langchainLib.get_llm("LLM.DEEPSEEK")
    # chain = RunnableParallel({
    #     "question": RunnablePassthrough(),
    #     "context": tool
    # }) | prompt | llm 
    # res = chain.invoke("易联众现在股价是多少？")
    # print(res)

    # print("#"*50)
    class Output(BaseModel):
        name:str = Field(description= "姓名")
        age:int = Field(description= "年龄")

    langchainLib.add_plugins()
    llm_key = args.llm_key
    llm = langchainLib.get_llm(llm_key)
    print("llms:",[(item["type"],item["api_key"],item["used"]) for item in langchainLib.get_llm(full=True)])

    outputParser = langchainLib.get_outputParser(Output,fix=True,llm = llm,retry=3)
    prompt = langchainLib.get_prompt(human_keys = {"ask":"问题"},use_chat=False)
    search = langchainLib.get_search_tool("TAVILY")
    tools = [search]
    #chain  = prompt | (lambda x: x.messages[-1].content) | search | (lambda x:'```json\n{"name":中国,"age":"20"}') | outputParser
    #chain  = prompt | llm | outputParser
    chain = prompt | RunnableLambda(lambda x:x.text,name="抽取text") | search 
    #print("chain to_json = ",chain.to_json())
    #print("chain name=",chain.get_name())
    print("\nchain graph=",chain.get_graph(),"\n")
    res = chain.invoke({"ask": "北京人口多少"})
    print(type(res),res)
    #FileLib.writeFile("graph1.png",chain.get_graph().draw_mermaid_png(),mode="wb")

    # llm = langchainLib.get_llm("LLM.DEEPSEEK")
    # print("llm",llm.to_json())
    llm.bind_tools(tools)
    res = llm.invoke({"ask":"北京2024年人口多少"})
    print(res)

def start(args):
    langchainLib = LangchainLib()
    langchainLib.add_plugins()
    if args.mode == "llm":
        StringLib.logging_in_box(f"\n{Color.YELLOW} 测试llm {Color.RESET}")
        __llm_test(langchainLib,args)
    elif args.mode == "prompt":
        StringLib.logging_in_box(f"\n{Color.YELLOW} 测试prompt {Color.RESET}")
        __prompt_test(langchainLib,args)    
    elif args.mode == 'loader':    
        StringLib.logging_in_box(f"\n{Color.YELLOW} 测试loader {Color.RESET}")
        start_loader(langchainLib,args)    
    elif args.mode == 'runnable':
        StringLib.logging_in_box(f"\n{Color.YELLOW} 测试runnable {Color.RESET}")
        __runnalble_test(langchainLib)
    elif args.mode == 'outputParser':    
        StringLib.logging_in_box(f"{Color.YELLOW} 测试outputParser {Color.RESET}")
        __outputParser_test(langchainLib,args)
    elif args.mode == 'rag':    
        StringLib.logging_in_box(f"\n{Color.YELLOW} 测试rag {Color.RESET}")
        start_rag(langchainLib,args)
    elif args.mode == 'tools':
        StringLib.logging_in_box(f"\n{Color.YELLOW} 测试tools {Color.RESET}")
        __tools_test(langchainLib,args)
    elif args.mode == 'graph':
        StringLib.logging_in_box(f"\n{Color.YELLOW} 测试graph {Color.RESET}")
        start_graph(langchainLib,args)
    elif args.mode == 'chat':
        start_chat(langchainLib,args)
    elif args.mode == 'agent':
        __agent(langchainLib,args)
    else:
        print("args=",args)
        print("llms--->:",[(item["type"],item["api_key"],item["model"],item["used"]) for item in langchainLib.get_llm(full=True)])
        print("embeddings---->:",[(item["type"],item["api_key"],item["model"],item["used"]) for item in langchainLib.get_embedding(full=True)])
        #loader = langchainLib.documentLib.pptx.loader("30335320.pptx")
        #docs = loader.load()
        #print(docs)
        
        # ## test elasticsearch vectorstore
        # langchainLib.vectorstoreLib.esLib.init_client(es_user="elastic")
        # langchainLib.vectorstoreLib.esLib.delete_store("langchain_index")
        # embed = langchainLib.embeddingLib.get_embedding('EMBEDDING.HF')
        # store = langchainLib.vectorstoreLib.esLib.get_store(embedding=embed)
        # ids = langchainLib.vectorstoreLib.esLib.create_from_texts(store,["我是一个科技工作者","树上有一只猴子"])
        # print(ids)
        # res = langchainLib.vectorstoreLib.esLib.search_with_score("我是谁？",store)
        # print(res)

        embed = langchainLib.embeddingLib.get_embedding(fake_size=10)
        print(embed.embed_documents(["hello world","I am fine!"]))

        return
        loader = langchainLib.documentLib.pptx.newer("test.pptx")
        loader.add_slide(0).set_title("Hello World","gogogo").add_text("youht",10,10,60,40)
        tab = [{"name":"youht","age":20},{"name":"jinli","age":10}] 
        tx = loader.add_slide(1).set_title("你好","step1").add_text("Step1",10,10,100,40)
        loader.add_text_paragraph(tx,"如何学习python",size=30,level=1)
        loader.add_text_paragraph(tx,"numpy",bold=True,size=20,level=2)
        loader.add_slide(2).set_title("你好","step1").add_table(tab,100,100,600,400,with_header = True)
        shape = loader.add_slide(8).set_title("Shape").add_shape(MSO_SHAPE.ROUNDED_RECTANGLE,50,50,50,50,
                                                        line_brightness=0.5,
                                                        text="hello")
        print(loader.get_shape_sizes(shape))
        loader.add_slide().set_title("chart1").add_chart("pie",
                                                        [{"s1":{"a":1,"b":2,"c":3}},
                                                         {"s2":{"a":3,"b":2,"c":1}}],10,10,600,300)
        loader.add_slide().set_title("chart2").add_chart("bubble",
                                                        [{"s1":[(1,2,10),(4,6,3),(2,3,1)]},
                                                         {"s2":[(2,1,2),(8,2,3),(4,1,5)]}],10,10,600,300)

        loader.save()


import re
import json
from typing import List
from bs4 import BeautifulSoup
from bs4.element import Tag, Comment,Doctype,Script,Stylesheet
from .crypto import HashLib
from .file import *    
import time
import requests
from lxml import etree
import uuid
class SoupLib():
    @classmethod
    def html2soup(cls, html) -> BeautifulSoup:
        #提取被```html包裹的部分，如果没有则转换整个html
        p=r"```html\n(.*?)\n```"
        res = re.findall(p,html,re.DOTALL)
        if res:
            html = res[0]
        #html = html.replace("<!DOCTYPE lake>","")
        soup = BeautifulSoup(html, "html.parser")
        
        #soup = BeautifulSoup(html, "html5lib")
        
        # 将标签外部的多个空白字符替换成一个空格
        # for element in soup.find_all(string=True):
        #     if element.parent.name not in ['script', 'style']:  # 忽略 <script> 和 <style> 标签内的内容
        #         element.replace_with(BeautifulSoup(" ".join(element.split()), 'html.parser'))
        return soup
    @classmethod
    def soup2html(cls, soup):
        #return soup.prettify()
        return soup.encode(formatter="html").decode('utf8')
    @classmethod
    def replace_block_with_tag(cls, soup, tag_name="keep", selectors=[]):
        #为符合selectors(xpath)的元素(所有)替换为<tag_name hash=<hash of elem>></tag_name>
        #如果已经有tag_name则不会重复嵌套tag_name
        #将elem保存在dictionary字典中，key为hash of elem
        # 返回新的soup 和 dictionary
        dictionary = {}
        tree = etree.HTML(str(soup))
        for selector in selectors:
            elems = tree.xpath(selector) 
            for elem in elems:
                hash = HashLib.md5(etree.tostring(elem).decode())[:6]
                dictionary[hash] = etree.tostring(elem).decode()
                tag_element = etree.Element(tag_name)
                tag_element.attrib["hash"] = hash
                parent = elem.getparent()
                if parent and parent.tag!=tag_name:
                    parent.replace(elem, tag_element)
        html = etree.tostring(tree).decode()
        return cls.html2soup(html),dictionary  

    @classmethod
    def restore_block_with_dict(cls, soup, restore_dict, tag_name="keep"):
        tree = etree.fromstring(str(soup))
        elems = tree.xpath(f"//{tag_name}") 
        for elem in elems:
            hash = elem.attrib["hash"]
            origin = restore_dict.get(hash)
            if origin:
                origin_elem = etree.fromstring(origin)
                parent = elem.getparent()
                parent.replace(elem,origin_elem)
        html = etree.tostring(tree).decode()
        return cls.html2soup(html)  

    @classmethod
    def wrap_block_with_tag(cls, soup, tag_name:str, selectors:List[str]) -> BeautifulSoup:
        #为符合selectors(xpath)的元素(第一个)套上tag_name
        #如果没有指定selectors，则直接wrap soup
        #如果已经有tag_name则不会重复嵌套tag_name
        # 返回新的soup
        tree = etree.HTML(str(soup))
        if selectors:
            for selector in selectors:
                elems = tree.xpath(selector) 
                if elems:
                    elem = elems[0]
                    tag_element = etree.Element(tag_name)
                    parent = elem.getparent()
                    if parent and parent.tag!=tag_name:
                        parent.replace(elem, tag_element)
                        tag_element.append(elem)
                        # parent = elem.getparent()
                        # index = parent.index(elem)
                        # parent.insert(index, tag_element)
            html = etree.tostring(tree).decode()
        else:
            tag_element = etree.Element(tag_name)
            tag_element.append(tree)
            html = etree.tostring(tag_element).decode()
        return cls.html2soup(html)            
    @classmethod
    def unwrap_block_with_tag(cls, soup, tag_name:str) -> BeautifulSoup:
        #去除tag_name
        #返回新的soup

        tree = etree.fromstring(str(soup))
        elems = tree.xpath(f"//{tag_name}") 
        for tag_element in elems:
            # 获取 <ignore> 标签的父元素
            parent = tag_element.getparent()
            # 获取 <ignore> 标签的索引
            index = parent.index(tag_element)
            # 将 <ignore> 标签的所有子元素移动到其父元素的相同索引位置
            for child in tag_element:
                tag_element.remove(child)
                parent.insert(index, child)
                index += 1
            # 移除 <ignore> 标签
            parent.remove(tag_element)
        html = etree.tostring(tree).decode()
        return cls.html2soup(html)  
    @classmethod
    def hash_attribute(cls, soup, key = 'hash',length=6):
        # 将soup中的所有标签的属性转化为hash，返回attribute_map且原soup已经变更
        attribute_dict = {}
        for tag in soup.find_all(True):  # 查找所有标签
            if tag.attrs:  # 如果标签有属性
                # 将属性字典转换为 "key1='value1' key2='value2'" 格式的字符串
                original_attributes = json.dumps(tag.attrs)

                # 计算 MD5 哈希值
                if length:
                    hash_value = HashLib.md5(original_attributes)[:length]
                else:
                    hash_value = HashLib.md5(original_attributes)

                # 保存原始属性到字典
                attribute_dict[hash_value] = original_attributes

                # 修改标签属性
                tag.attrs.clear()  # 清除所有现有属性
                tag.attrs[key] = hash_value

        return attribute_dict
    @classmethod
    def unhash_attribute(cls,soup, attribute_dict, key = 'hash'):
        # 根据attribute_map还原soup
        for tag in soup.find_all(True):
            if key in tag.attrs:
                hash_value = tag.attrs[key]
                if hash_value in attribute_dict:
                    # 从字典中获取原始属性
                    original_attributes = json.loads(attribute_dict[hash_value])

                    # 还原原始属性
                    tag.attrs.clear()  # 清除现有属性
                    tag.attrs.update(original_attributes)
    @classmethod
    def compare_soup_structure(cls, soup1, soup2):
        """
        比较两个 BeautifulSoup 对象的结构是否完全一致，忽略文本内容。

        Args:
            soup1: 第一个 BeautifulSoup 对象。
            soup2: 第二个 BeautifulSoup 对象。

        Returns:
            bool: 如果结构完全一致则返回 True，否则返回 False。
        """

        def compare_tags(tag1, tag2):
            """
            递归比较两个标签及其子元素的结构，忽略文本内容。

            Args:
                tag1: 第一个标签。
                tag2: 第二个标签。

            Returns:
                bool: 如果结构完全一致则返回 True，否则返回 False。
            """
            if tag1.name != tag2.name:
                return False
            if tag1.attrs != tag2.attrs:
                return False
            if len(tag1.contents) != len(tag2.contents):
                return False
            for i in range(len(tag1.contents)):
                child1 = tag1.contents[i]
                child2 = tag2.contents[i]
                if isinstance(child1, Tag) and isinstance(child2, Tag):
                    if not compare_tags(child1, child2):
                        return False
            return True

        return compare_tags(soup1, soup2)
    @classmethod
    def mask_text_with_dictionary(cls, soup, dictionary={},key="__m__"):
        for element in soup.find_all(text=True):
            if element.parent.name in ['[document]', 'html', 'body']:
                continue  # 跳过顶层元素的文本节点
            if element.strip() and not element.string:
                continue  # 跳过包含子元素的文本节点

            # 尝试替换位mask：hash=******
            text = element.strip()
            hash = HashLib.md5(text)[:6]
            new_text = dictionary.get(hash,{}).get("target_text")
            if new_text:
                element.replace_with(f"{key}={hash}")
    @classmethod
    def mask_html_with_dictionary(cls, soup, attrs: list[str], value_key = "value", dictionary={},key="__m__"):
        #对于soup中所有含有全部attrs属性的元素，如果其value_key的hash值在dictionary中有值，则将该元素的内容mask为__m__=<value_hash>
        def _hasattrs(tag):
            has = True
            for attr in attrs:
                if not tag.get(attr):
                    return False
            return True
        for idx,element in enumerate(soup.find_all(lambda tag: _hasattrs(tag))):
            value_hash = element.attrs[value_key]
            new_text = dictionary.get(value_hash,{}).get("target_text")
            logging.debug(f"mask_html_with_dictionary---> idx={idx},value_hash={value_hash},bool(new_text)={bool(new_text)}")
            if new_text:
                element.string = f"{key}={value_hash}"
    @classmethod
    def unmask_html_with_dictionary(cls, soup, attrs:list[str],value_key = "value", dictionary={},key="__m__"):
        def _hasattrs(tag):
            has = True
            for attr in attrs:
                if not tag.get(attr):
                    return False
            return True
        for element in soup.find_all(lambda tag: _hasattrs(tag)):
            # 尝试替换文本
            hashs = re.findall(f"{key}=(.{{6}})","".join([str(el) for el in element.contents]))
            if hashs:
                new_text = dictionary.get(hashs[0],{}).get("target_text")
                if new_text:
                    element.clear()
                    element.append(SoupLib.html2soup(new_text))
    @classmethod
    def unmask_text_with_dictionary(cls, soup, dictionary={},key="__m__"):
        for element in soup.find_all(text=True):
            if element.parent.name in ['[document]', 'html', 'body']:
                continue  # 跳过顶层元素的文本节点
            if element.strip() and not element.string:
                continue  # 跳过包含子元素的文本节点

            
            # 尝试替换文本
            text = element.strip()
            hashs = re.findall(f"{key}=(.{{6}})",text)
            if hashs:
                new_text = dictionary.get(hashs[0],{}).get("target_text")
                if new_text:
                    element.replace_with(new_text)
    @classmethod
    def replace_text_with_dictionary(cls, soup, dictionary={}):
        for element in soup.find_all(text=True):
            if element.parent.name in ['[document]', 'html', 'body']:
                continue  # 跳过顶层元素的文本节点
            if element.strip() and not element.string:
                continue  # 跳过包含子元素的文本节点

            # 尝试替换文本
            new_text = dictionary.get(element.strip())
            if new_text:
                element.replace_with(new_text)
    @classmethod
    def find_all_text(cls, soup,separator='_||_')->list:
        texts = soup.get_text(separator=separator, strip=True)
        if not texts:
            return []
        return texts.split(separator)
    @classmethod
    def find_all_text_without_mask(cls, soup,separator='_||_',mask_key='__m__')->list:
        texts = soup.get_text(separator=separator, strip=True)
        if not texts:
            return []
        texts_list = [text for text in texts.split(separator) if not re.match(f"{mask_key}=(.{{6}})",text)]
        return texts_list

    @classmethod
    def walk(cls, soup, func=None,size=200,level=0,blocks=[],ignore_tags=["script","style","ignore"],value_dict={}):
        contents = soup.contents
        length = 0
        nodes = {}
        for idx,node in enumerate(contents):
            is_ignored = False
            if node.name in ignore_tags:
                is_ignored = True
            #print("ignore=",is_ignored,"node.name=",node.name,"level=",level,"idx=",idx,"length=",length,"node=",len(str(node)),str(node)[:50],"...")        
            if not is_ignored:
                length += len(str(node))
                if length < size:
                    hash = HashLib.md5(str(uuid.uuid4()))[:6]
                    nodes[hash]=node
                    continue
            length = 0
            if nodes:
                #print("level=",level,f"replace with block {len(blocks)}")
                nodes_html=""
                for key in nodes:
                    if isinstance(nodes[key],Doctype):
                        node_html = f"<!DOCTYPE {str(nodes[key])}>"
                    elif isinstance(nodes[key],Tag):
                        node_html = str(nodes[key])
                    else:
                        node_html = str(nodes[key])
                        #print("*"*80,type(nodes[key]),"\n\n","@",str(nodes[key]),"@","*"*80)
                    value_hash = HashLib.md5(node_html)[:6]
                    value_dict[value_hash] = node_html
                    cls.replace_block(nodes[key],BeautifulSoup(f"<div path={key} value={value_hash}></div>", 'html.parser'))
                    nodes_html += f"<div path={key} value={value_hash}>{node_html}</div>"                
                blocks.append(nodes_html)
                nodes = {}

            if not is_ignored:
                if len(str(node)) < size:
                    hash = HashLib.md5(str(uuid.uuid4()))[:6]
                    nodes[hash]=node
                    length += len(str(node))
                    continue
                else:
                    if isinstance(node, Tag): 
                        cls.walk(node, func=func,size=size,level=level+1,blocks=blocks,ignore_tags=ignore_tags,value_dict=value_dict)
                    else:
                        func and  func(node)
        if nodes:
                #print("level=",level,f"replace with block {len(blocks)}")
                nodes_html=""
                for key in nodes:
                    if isinstance(nodes[key],Doctype):
                        node_html = f"<!DOCTYPE {str(nodes[key])}>"
                    elif isinstance(nodes[key],Tag):
                        node_html = str(nodes[key])
                    else:
                        node_html = str(nodes[key])
                    value_hash = HashLib.md5(node_html)[:6]
                    value_dict[value_hash] = node_html
                    cls.replace_block(nodes[key],BeautifulSoup(f"<div path={key} value={value_hash}></div>", 'html.parser'))
                    nodes_html += f"<div path={key} value={value_hash}>{node_html}</div>"
                blocks.append(nodes_html)
                nodes = {}
    @classmethod
    def replace_block(cls, source_block, target_block):
        source_block.replace_with(target_block)
    @classmethod
    def unwalk(cls,soup,blocks):
        for idx, block in enumerate(blocks):
            target_block = cls.html2soup(block)
            tnodes = target_block.find_all(lambda tag:tag.has_attr("path") and tag.has_attr("value"))
            for target_node in tnodes:
                t = target_node.attrs.get("path")
                source_node = soup.find(attrs={"path":t})
                if source_node:
                    #这种方法的问题是target_node.contents可能包含前后两个"\n",导致真正的元素丢失,该用扩展target_node.contents,然后unwrap
                    #print("idx=",idx,"t=",t,"target_node.contents[0]=",len(target_node.contents),[str(item) for item in target_node.contents])
                    #source_node.replace_with(target_node.contents[0])
                    source_node.extend(target_node.contents)
                    source_node.unwrap()
                    #FileLib.writeFile(f"{idx}-{t}.html",SoupLib.soup2html(soup))
    @classmethod
    def replace_tag_name(cls,soup,old_tag_name,new_tag_name):
        old_tags = [tag for tag in soup.find_all() if tag.name == old_tag_name]
        for old_tag in old_tags:
            new_tag = soup.new_tag(new_tag_name)
            for attr, value in old_tag.attrs.items():
                new_tag[attr] = value            
            new_tag.extend(old_tag.contents)
            old_tag.replace_with(new_tag)

if __name__ == "__main__":
    html0 = requests.get("https://global.alipay.com/docs/").text
    html =  FileLib.readFile("pw.html")
    html2 = """
<body>
  <div>
    hello world!!
    厦门
    漳州
    泉州
    <h1> 对话 <em> 重点 </em>   </h1>

    
    <p> let's go to home </p>
    <p> ok, that is fine </p>
  </div>
  <div>
    <h1> 其他测试 </h1>
    <div> 
    </div>
    <div> <p> I am a student!</p></div>
  </div>
  <script>
    var a = 1;
    var b = 2;

  </script>
</body>
"""
    soup = SoupLib.html2soup(html)
    hash_dict = SoupLib.hash_attribute(soup)
    blocks = []
    SoupLib.walk(soup, size=800,blocks=blocks)
    print("="*20)
    #print(soup)
    FileLib.writeFile("soup1.html",SoupLib.soup2html(soup))
    print("="*50)
    for idx,block in enumerate(blocks):
        print(idx,block)

    print("="*80)
    SoupLib.unwalk(soup,blocks)
    SoupLib.unhash_attribute(soup,hash_dict)
    #print(soup.prettify())
    FileLib.writeFile("soup2.html",SoupLib.soup2html(soup))






from langchain_core.tools import tool
from typing import Optional
import sqlite3

from ylz_utils.cli.custom_service.db_init import db

@tool
def search_trip_recommendations(
    location: Optional[str] = None,
    name: Optional[str] = None,
    keywords: Optional[str] = None,
) -> list[dict]:
    """
    Search for trip recommendations based on location, name, and keywords.

    Args:
        location (Optional[str]): The location of the trip recommendation. Defaults to None.
        name (Optional[str]): The name of the trip recommendation. Defaults to None.
        keywords (Optional[str]): The keywords associated with the trip recommendation. Defaults to None.

    Returns:
        list[dict]: A list of trip recommendation dictionaries matching the search criteria.
    """
    conn = sqlite3.connect(db)
    cursor = conn.cursor()

    query = "SELECT * FROM trip_recommendations WHERE 1=1"
    params = []

    if location:
        query += " AND location LIKE ?"
        params.append(f"%{location}%")
    if name:
        query += " AND name LIKE ?"
        params.append(f"%{name}%")
    if keywords:
        keyword_list = keywords.split(",")
        keyword_conditions = " OR ".join(["keywords LIKE ?" for _ in keyword_list])
        query += f" AND ({keyword_conditions})"
        params.extend([f"%{keyword.strip()}%" for keyword in keyword_list])

    cursor.execute(query, params)
    results = cursor.fetchall()

    conn.close()

    return [
        dict(zip([column[0] for column in cursor.description], row)) for row in results
    ]


@tool
def book_excursion(recommendation_id: int) -> str:
    """
    Book a excursion by its recommendation ID.

    Args:
        recommendation_id (int): The ID of the trip recommendation to book.

    Returns:
        str: A message indicating whether the trip recommendation was successfully booked or not.
    """
    conn = sqlite3.connect(db)
    cursor = conn.cursor()

    cursor.execute(
        "UPDATE trip_recommendations SET booked = 1 WHERE id = ?", (recommendation_id,)
    )
    conn.commit()

    if cursor.rowcount > 0:
        conn.close()
        return f"Trip recommendation {recommendation_id} successfully booked."
    else:
        conn.close()
        return f"No trip recommendation found with ID {recommendation_id}."


@tool
def update_excursion(recommendation_id: int, details: str) -> str:
    """
    Update a trip recommendation's details by its ID.

    Args:
        recommendation_id (int): The ID of the trip recommendation to update.
        details (str): The new details of the trip recommendation.

    Returns:
        str: A message indicating whether the trip recommendation was successfully updated or not.
    """
    conn = sqlite3.connect(db)
    cursor = conn.cursor()

    cursor.execute(
        "UPDATE trip_recommendations SET details = ? WHERE id = ?",
        (details, recommendation_id),
    )
    conn.commit()

    if cursor.rowcount > 0:
        conn.close()
        return f"Trip recommendation {recommendation_id} successfully updated."
    else:
        conn.close()
        return f"No trip recommendation found with ID {recommendation_id}."


@tool
def cancel_excursion(recommendation_id: int) -> str:
    """
    Cancel a trip recommendation by its ID.

    Args:
        recommendation_id (int): The ID of the trip recommendation to cancel.

    Returns:
        str: A message indicating whether the trip recommendation was successfully cancelled or not.
    """
    conn = sqlite3.connect(db)
    cursor = conn.cursor()

    cursor.execute(
        "UPDATE trip_recommendations SET booked = 0 WHERE id = ?", (recommendation_id,)
    )
    conn.commit()

    if cursor.rowcount > 0:
        conn.close()
        return f"Trip recommendation {recommendation_id} successfully cancelled."
    else:
        conn.close()
        return f"No trip recommendation found with ID {recommendation_id}."


import sqlite3
from langchain_core.tools import tool 
from typing import Optional,Union
from datetime import datetime, date

from ylz_utils.cli.custom_service.db_init import db
@tool
def search_hotels(
    location: Optional[str] = None,
    name: Optional[str] = None,
    price_tier: Optional[str] = None,
    checkin_date: Optional[Union[datetime, date]] = None,
    checkout_date: Optional[Union[datetime, date]] = None,
) -> list[dict]:
    """
    Search for hotels based on location, name, price tier, check-in date, and check-out date.

    Args:
        location (Optional[str]): The location of the hotel. Defaults to None.
        name (Optional[str]): The name of the hotel. Defaults to None.
        price_tier (Optional[str]): The price tier of the hotel. Defaults to None. Examples: Midscale, Upper Midscale, Upscale, Luxury
        checkin_date (Optional[Union[datetime, date]]): The check-in date of the hotel. Defaults to None.
        checkout_date (Optional[Union[datetime, date]]): The check-out date of the hotel. Defaults to None.

    Returns:
        list[dict]: A list of hotel dictionaries matching the search criteria.
    """
    conn = sqlite3.connect(db)
    cursor = conn.cursor()

    query = "SELECT * FROM hotels WHERE 1=1"
    params = []

    if location:
        query += " AND location LIKE ?"
        params.append(f"%{location}%")
    if name:
        query += " AND name LIKE ?"
        params.append(f"%{name}%")
    # For the sake of this tutorial, we will let you match on any dates and price tier.
    cursor.execute(query, params)
    results = cursor.fetchall()

    conn.close()

    return [
        dict(zip([column[0] for column in cursor.description], row)) for row in results
    ]


@tool
def book_hotel(hotel_id: int) -> str:
    """
    Book a hotel by its ID.

    Args:
        hotel_id (int): The ID of the hotel to book.

    Returns:
        str: A message indicating whether the hotel was successfully booked or not.
    """
    conn = sqlite3.connect(db)
    cursor = conn.cursor()

    cursor.execute("UPDATE hotels SET booked = 1 WHERE id = ?", (hotel_id,))
    conn.commit()

    if cursor.rowcount > 0:
        conn.close()
        return f"Hotel {hotel_id} successfully booked."
    else:
        conn.close()
        return f"No hotel found with ID {hotel_id}."


@tool
def update_hotel(
    hotel_id: int,
    checkin_date: Optional[Union[datetime, date]] = None,
    checkout_date: Optional[Union[datetime, date]] = None,
) -> str:
    """
    Update a hotel's check-in and check-out dates by its ID.

    Args:
        hotel_id (int): The ID of the hotel to update.
        checkin_date (Optional[Union[datetime, date]]): The new check-in date of the hotel. Defaults to None.
        checkout_date (Optional[Union[datetime, date]]): The new check-out date of the hotel. Defaults to None.

    Returns:
        str: A message indicating whether the hotel was successfully updated or not.
    """
    conn = sqlite3.connect(db)
    cursor = conn.cursor()

    if checkin_date:
        cursor.execute(
            "UPDATE hotels SET checkin_date = ? WHERE id = ?", (checkin_date, hotel_id)
        )
    if checkout_date:
        cursor.execute(
            "UPDATE hotels SET checkout_date = ? WHERE id = ?",
            (checkout_date, hotel_id),
        )

    conn.commit()

    if cursor.rowcount > 0:
        conn.close()
        return f"Hotel {hotel_id} successfully updated."
    else:
        conn.close()
        return f"No hotel found with ID {hotel_id}."


@tool
def cancel_hotel(hotel_id: int) -> str:
    """
    Cancel a hotel by its ID.

    Args:
        hotel_id (int): The ID of the hotel to cancel.

    Returns:
        str: A message indicating whether the hotel was successfully cancelled or not.
    """
    conn = sqlite3.connect(db)
    cursor = conn.cursor()

    cursor.execute("UPDATE hotels SET booked = 0 WHERE id = ?", (hotel_id,))
    conn.commit()

    if cursor.rowcount > 0:
        conn.close()
        return f"Hotel {hotel_id} successfully cancelled."
    else:
        conn.close()
        return f"No hotel found with ID {hotel_id}."


import logging
import os
import shutil

def copy_config(args):
    config_name = args.config_name
    project_name = args.project_name
    current_path = os.path.abspath(__file__)
    config_dir = os.path.dirname(current_path)
    config_file = os.path.join(config_dir,"..",config_name)
    dest_dir = os.path.join(os.path.expanduser("~"), f".{project_name}")
    if not os.path.exists(dest_dir):
        os.makedirs(dest_dir)
    target_file = os.path.join(dest_dir, config_name)
    if not os.path.exists(target_file):
        shutil.copy(config_file, dest_dir)
    else:
        logging.info(f"已存在的{config_name}将被覆盖.")
        os.remove(target_file)
        shutil.copy(config_file, dest_dir)

def reset(args):
    copy_config(args)



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib
    
from langchain_core.prompts import ChatPromptTemplate,PromptTemplate,SystemMessagePromptTemplate
from langchain_core.prompt_values import StringPromptValue,ChatPromptValue

from ylz_utils.data import StringLib

class PromptLib():
    def get_prompt(self,system_prompt=None,human_keys={"input":""},outputParser=None,history_messages_key="history",use_chat = False,use_chinese=True) -> ChatPromptTemplate:
            if not system_prompt:
                system_prompt=""
            if use_chinese:
                system_prompt = f"所有问题请用中文回答\n{system_prompt}"
            if not use_chat:
                human_input_keys = []
                if human_keys:
                    human_prompt = ""
                    for key in human_keys:
                        human_prompt += f"{human_keys[key]}:{{{key}}}\n" 
                    human_input_keys = human_keys.keys()
                if outputParser:
                    prompt = PromptTemplate(
                        template=f"{system_prompt}\n{{format_instructions}}\n{human_prompt}",
                        input_variables=human_input_keys,
                        partial_variables={"format_instructions": outputParser.get_format_instructions()}
                    )
                else:
                    prompt =  PromptTemplate(
                        template=f"{system_prompt}\n{human_prompt}",
                        input_variables=human_input_keys,
                    ) 
            else:
                messages = []
                if outputParser:
                    partial_prompt = PromptTemplate(template=f"{system_prompt}\n{{format_instructions}}", 
                                                    partial_variables={"format_instructions": outputParser.get_format_instructions()})
                    #messages.append(("system",partial_prompt.format(**{"format_instructions": outputParser.get_format_instructions() })))
                    messages.append(
                        SystemMessagePromptTemplate(prompt=partial_prompt)               
                    )
                elif system_prompt:
                    messages.append(("system",system_prompt))

                messages.append(("placeholder", f"{{{history_messages_key}}}"))
                
                if human_keys:
                    human_prompt = ""
                    for key in human_keys:
                        human_prompt += f"{human_keys[key]}:{{{key}}}\n"
                    messages.append(("human",human_prompt))
                StringLib.logging_in_box(str(messages),console_width=160,print_func=print)
                prompt = ChatPromptTemplate.from_messages(messages)
            return prompt



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib

from langgraph.prebuilt import chat_agent_executor

class AgentLib():
    def __init__(self,langchainLib:LangchainLib):
        self.langchainLib = langchainLib
    def get_agent(self,llm,tools):
        return chat_agent_executor.create_function_calling_executor(llm,tools)



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib

import random
import re

from langchain_community.utilities.tavily_search import TavilySearchAPIWrapper
from langchain_community.tools.tavily_search.tool import TavilySearchResults
from langchain_community.utilities.duckduckgo_search import DuckDuckGoSearchAPIWrapper
from langchain_community.tools import DuckDuckGoSearchResults

from langchain.docstore.document import Document
from langchain_core.runnables import RunnableLambda

class WebSearchTool():
    def __init__(self,langchainLib:LangchainLib):
        self.langchainLib = langchainLib
        self.config = langchainLib.config
    def get_tool_wrapper(self):
        search_config = self.config.get(f"SEARCH_TOOLS.TAVILY")        
        api_keys = search_config.get("API_KEYS")
        try:
            if api_keys:
                api_keys = api_keys.split(",")
            else:
                api_keys = []
            api_key = random.choice(api_keys)
        except:
            raise Exception(f"请先设置TAVILY_API_KEYS环境变量") 
        search = TavilySearchAPIWrapper(tavily_api_key=api_key)
        tool = TavilySearchResults(api_wrapper=search,max_results=4)
        return tool

    def get_tool(self,key="TAVILY",rows=4):
        key = key.upper()
        if key == "DUCKDUCKGO":
            search = DuckDuckGoSearchAPIWrapper()
            tool  = DuckDuckGoSearchResults(api_wrapper=search,num_results=rows)
            #snippet,title,link: 
            pattern = "snippet: (.*?) title: (.*?) link: (.*?) snippet:"
            def __toDocument(text):
                pattern = r"\[snippet: (.*?) title: (.*?) link: (.*?)\]"
                matchs = re.findall(pattern,text)
                docs = [Document(match[0],metadata={"title":match[1],"link":match[2]}) for match in matchs]
                return docs
            tool_doc = tool | RunnableLambda(__toDocument, name="DDG2Document")
            #return tool_doc
            return tool 
        elif key == "TAVILY":
            # url,content,
            search_config = self.config.get(f"SEARCH_TOOLS.{key}")        
            api_keys = search_config.get("API_KEYS")
            try:
                api_keys = self.langchainLib.split_keys(api_keys)
                api_key = random.choice(api_keys)
            except:
                raise Exception(f"请先设置{key}_API_KEYS环境变量") 
            if api_key:
                try:
                    def __toDocument(doc_json):
                        docs = [Document(doc['content'],metadata={"link":doc['url']}) for doc in doc_json]
                        return docs
                    search = TavilySearchAPIWrapper(tavily_api_key=api_key)
                    tool = TavilySearchResults(api_wrapper=search,max_results=rows)
                    tool_doc = tool | RunnableLambda(__toDocument, name="Tavily2Document")
                    #return tool_doc
                    return tool
                except:
                    raise Exception("请先设置TAVILY_API_KEYS环境变量") 
        else:
            raise Exception(f"不支持{key}的搜索") 



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib

import random
import time
from typing import Optional

from ylz_utils import Config

from langchain_together import TogetherEmbeddings
from langchain_google_genai import GoogleGenerativeAIEmbeddings
from langchain_ollama import OllamaEmbeddings

from langchain_huggingface import HuggingFaceEndpointEmbeddings,HuggingFaceEmbeddings

from langchain_core.embeddings import DeterministicFakeEmbedding,Embeddings

class EmbeddingLib():
    embeddings:list  = []
    default_embedding_key = None
    fake_size:int = 0
    def __init__(self,langchainLib:LangchainLib):
        self.langchainLib = langchainLib
        self.config = Config.get()
        self.regist_embedding()
        if self.config.get("EMBEDDING.DEFAULT"):
            default_embedding_key = self.config.get("EMBEDDING.DEFAULT")
            self.set_default_embedding_key(default_embedding_key)
        else:
            self.clear_default_embedding_key()
    def set_default_embedding_key(self,key):
        embeddings = [item for item in self.embeddings if item['type']==key]
        if embeddings:
            self.default_embedding_key = key
        else:
            self.clear_default_embedding_key()
    def get_default_embedding_key(self):
        return self.default_embedding_key
    def clear_default_embedding_key(self):
        self.default_embedding_key=None
    
    def set_fake_size(self,size=0):
        if size:
            self.fake_size = size
        else:
            self.fake_size = 0
    def get_embedding(self,key=None,model=None, full=False, fake_size: Optional[int]=None) :
        if full:
            return self.embeddings
        if fake_size:
            return DeterministicFakeEmbedding(size = fake_size)
        if self.fake_size:
            return DeterministicFakeEmbedding(size = self.fake_size)
        if self.embeddings:
            if not key:
                if self.default_embedding_key:
                    embeddings = [item for item in self.embeddings if item['type']==self.default_embedding_key]
                else:
                    embeddings = self.embeddings
            else:
                embeddings = [item for item in self.embeddings if item['type']==key]
            key_sets = set([item['type'] for item in embeddings])
            if len(key_sets) > 1:
                raise Exception(f"请确保可选的llm_key({key_sets})只有一种,然后调用regist_llm注册语言模型")
            if embeddings:
                #embedding = random.choice(embeddings)    
                embedding = embeddings[0]
                if not embedding.get('embedding'):
                    embed_type = embedding['type']
                    if  embed_type == 'EMBEDDING.TOGETHER':
                        embedding['embedding'] = TogetherEmbeddings(api_key = embedding.get('api_key'),model=embedding.get('model'))
                    elif embed_type == "EMBEDDING.GEMINI" :
                        embedding['embedding'] = GoogleGenerativeAIEmbeddings(model=embedding.get('model'),google_api_key=embedding.get('api_key'))
                    elif embed_type == 'EMBEDDING.OLLAMA':
                        embedding['embedding'] = OllamaEmbeddings(model=embedding.get('model'))
                    elif embed_type == 'EMBEDDING.HF':
                        embedding['embedding'] = HuggingFaceEmbeddings(model_name=embedding.get('model'),
                                                                       model_kwargs = {'device': 'cpu'},
                                                                       encode_kwargs = {'normalize_embeddings': False})
                    else:
                        raise Exception(f"目前不支持{embedding['type']}嵌入模型")
                embedding['used'] = embedding.get('used',0) + 1 
                embedding['last_used'] = time.time()
                return embedding['embedding']
        if key:
            raise Exception(f"请确保{key}_API_KEYS环境变量被正确设置,然后调用regist_embedding注册语言模型")
        if self.default_embedding_key:
            raise Exception(f"请确保{self.default_embedding_key}_API_KEYS环境变量被正确设置,然后调用regist_embedding注册语言模型")
        else:
            raise Exception(f"请确保<EMBEDDING PROVIDER>_API_KEYS环境变量被正确设置,然后调用regist_embedding注册语言模型")
    
    def regist_embedding(self):
        defaults = {
                      "EMBEDDING.TOGETHER": {"model":"BAAI/bge-large-en-v1.5"},
                      "EMBEDDING.GEMINI": {"model":"models/embedding-001"},
                      "EMBEDDING.OLLAMA": {"model":"mxbai-embed-large"},
                      "EMBEDDING.HF": {"model":"BAAI/bge-large-en"}  #"sentence-transformers/all-mpnet-base-v2"
                  }
        for key in defaults:
            default = defaults[key]
            embed = self.config.get(key)
            if not embed:
                continue
            base_url = embed.get("BASE_URL")
            api_keys = embed.get("API_KEYS")
            api_keys = self.langchainLib.split_keys(api_keys)

            model= embed.get("MODEL") if embed.get("MODEL") else default['model']
            for api_key in api_keys:
                self.embeddings.append({
                    "embedding": None,
                    "type": key,
                    "base_url": base_url,
                    "api_key":api_key,
                    "model":model,
                    "used":0,
                    "last_used": 0 
                })
    
    def embed_documents(self,embed:Embeddings,texts:list[str]):
        return embed.embed_documents(texts)
    
    def embed_query(self,embed:Embeddings,text:str):
        return embed.embed_query(text)
    



from __future__ import annotations
from typing import TYPE_CHECKING
if TYPE_CHECKING:
    from ylz_utils.langchain import LangchainLib

from langchain_community.document_loaders import RecursiveUrlLoader
from langchain_community.document_transformers import MarkdownifyTransformer
from langchain_core.documents import Document

class UrlLib():
    def __init__(self,langchainLib:LangchainLib):
        self.langchainLib = langchainLib
    def loader(self, url, max_depth=2, extractor=None, metadata_extractor=None):
        #"./example_data/fake.docx"
        loader = RecursiveUrlLoader(
            url = url,
            max_depth= max_depth,
            # use_async=False,
            extractor= extractor,
            metadata_extractor= metadata_extractor
            # exclude_dirs=(),
            # timeout=10,
            # check_response_status=True,
            # continue_on_failure=True,
            # prevent_outside=True,
            # base_url=None,
        )
        return loader

    def load_and_split_markdown(self, url, max_depth=2, extractor=None, metadata_extractor=None, chunk_size=1000,chunk_overlap=0):
        loader = self.loader(url,max_depth=max_depth,extractor=extractor,metadata_extractor=metadata_extractor)
        docs = loader.load()
        transformer = MarkdownifyTransformer()
        converted_docs = transformer.transform_documents(docs)
        result = []
        for doc in converted_docs:
            splited_docs = self.langchainLib.splitterLib.split_markdown_docs(doc.page_content,chunk_size=chunk_size,chunk_overlap=chunk_overlap)
            result.append({"doc":doc,"blocks":splited_docs,"metadata":doc.metadata})
        return result
    
    def load_and_split(self,url,max_depth=2, extractor=None, metadata_extractor=None, chunk_size=1000,chunk_overlap=0):
        res = self.load_and_split_markdown(url,mac_depth=max_depth,
                                           extractor=extractor,
                                           metadata_extractor=metadata_extractor,
                                           chunk_size = chunk_size,
                                           chunk_overlap = chunk_overlap)
        docs = []
        for doc in res:
            for block in doc["blocks"]:
                docs.append(Document(block.page_content,metadata=doc["metadata"]))
        return docs


from datetime import date, datetime
from typing import Optional, Union

from langchain_core.tools import tool
import sqlite3

from ylz_utils.cli.custom_service.db_init import db

@tool
def search_car_rentals(
    location: Optional[str] = None,
    name: Optional[str] = None,
    price_tier: Optional[str] = None,
    start_date: Optional[Union[datetime, date]] = None,
    end_date: Optional[Union[datetime, date]] = None,
) -> list[dict]:
    """
    Search for car rentals based on location, name, price tier, start date, and end date.

    Args:
        location (Optional[str]): The location of the car rental. Defaults to None.
        name (Optional[str]): The name of the car rental company. Defaults to None.
        price_tier (Optional[str]): The price tier of the car rental. Defaults to None.
        start_date (Optional[Union[datetime, date]]): The start date of the car rental. Defaults to None.
        end_date (Optional[Union[datetime, date]]): The end date of the car rental. Defaults to None.

    Returns:
        list[dict]: A list of car rental dictionaries matching the search criteria.
    """
    conn = sqlite3.connect(db)
    cursor = conn.cursor()

    query = "SELECT * FROM car_rentals WHERE 1=1"
    params = []

    if location:
        query += " AND location LIKE ?"
        params.append(f"%{location}%")
    if name:
        query += " AND name LIKE ?"
        params.append(f"%{name}%")
    # For our tutorial, we will let you match on any dates and price tier.
    # (since our toy dataset doesn't have much data)
    cursor.execute(query, params)
    results = cursor.fetchall()

    conn.close()

    return [
        dict(zip([column[0] for column in cursor.description], row)) for row in results
    ]


@tool
def book_car_rental(rental_id: int) -> str:
    """
    Book a car rental by its ID.

    Args:
        rental_id (int): The ID of the car rental to book.

    Returns:
        str: A message indicating whether the car rental was successfully booked or not.
    """
    conn = sqlite3.connect(db)
    cursor = conn.cursor()

    cursor.execute("UPDATE car_rentals SET booked = 1 WHERE id = ?", (rental_id,))
    conn.commit()

    if cursor.rowcount > 0:
        conn.close()
        return f"Car rental {rental_id} successfully booked."
    else:
        conn.close()
        return f"No car rental found with ID {rental_id}."


@tool
def update_car_rental(
    rental_id: int,
    start_date: Optional[Union[datetime, date]] = None,
    end_date: Optional[Union[datetime, date]] = None,
) -> str:
    """
    Update a car rental's start and end dates by its ID.

    Args:
        rental_id (int): The ID of the car rental to update.
        start_date (Optional[Union[datetime, date]]): The new start date of the car rental. Defaults to None.
        end_date (Optional[Union[datetime, date]]): The new end date of the car rental. Defaults to None.

    Returns:
        str: A message indicating whether the car rental was successfully updated or not.
    """
    conn = sqlite3.connect(db)
    cursor = conn.cursor()

    if start_date:
        cursor.execute(
            "UPDATE car_rentals SET start_date = ? WHERE id = ?",
            (start_date, rental_id),
        )
    if end_date:
        cursor.execute(
            "UPDATE car_rentals SET end_date = ? WHERE id = ?", (end_date, rental_id)
        )

    conn.commit()

    if cursor.rowcount > 0:
        conn.close()
        return f"Car rental {rental_id} successfully updated."
    else:
        conn.close()
        return f"No car rental found with ID {rental_id}."


@tool
def cancel_car_rental(rental_id: int) -> str:
    """
    Cancel a car rental by its ID.

    Args:
        rental_id (int): The ID of the car rental to cancel.

    Returns:
        str: A message indicating whether the car rental was successfully cancelled or not.
    """
    conn = sqlite3.connect(db)
    cursor = conn.cursor()

    cursor.execute("UPDATE car_rentals SET booked = 0 WHERE id = ?", (rental_id,))
    conn.commit()

    if cursor.rowcount > 0:
        conn.close()
        return f"Car rental {rental_id} successfully cancelled."
    else:
        conn.close()
        return f"No car rental found with ID {rental_id}."